import argparse
import typing
from pathlib import Path

import fiona
import geopandas as gpd
import numpy as np
import shapely

from geograypher.cameras import MetashapeCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.meshes import TexturedPhotogrammetryMesh
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshChunked
from geograypher.utils.visualization import show_segmentation_labels


# TODO Consider adding a mesh screenshot option or a vis savefolder
def render_labels(
    mesh_file: PATH_TYPE,
    cameras_file: PATH_TYPE,
    image_folder: PATH_TYPE,
    texture: typing.Union[PATH_TYPE, np.ndarray, None],
    render_savefolder: PATH_TYPE,
    transform_file: typing.Union[PATH_TYPE, None] = None,
    subset_images_savefolder: typing.Union[PATH_TYPE, None] = None,
    texture_column_name: typing.Union[str, None] = None,
    DTM_file: typing.Union[PATH_TYPE, None] = None,
    ground_height_threshold: typing.Union[float, None] = None,
    render_ground_class: bool = False,
    textured_mesh_savefile: typing.Union[PATH_TYPE, None] = None,
    ROI: typing.Union[PATH_TYPE, gpd.GeoDataFrame, shapely.MultiPolygon, None] = None,
    mesh_ROI_buffer_radius_meters: float = 50,
    cameras_ROI_buffer_radius_meters: float = 150,
    IDs_to_labels: typing.Union[dict, None] = None,
    render_image_scale: float = 1,
    mesh_downsample: float = 1,
    n_render_clusters: typing.Union[int, None] = None,
    vis: bool = False,
    mesh_vis_file: typing.Union[PATH_TYPE, None] = None,
    labels_vis_folder: typing.Union[PATH_TYPE, None] = None,
):
    """Renders image-based labels using geospatial ground truth data

    Args:
        mesh_file (PATH_TYPE):
            Path to the Metashape-exported mesh file
        cameras_file (PATH_TYPE):
            Path to the MetaShape-exported .xml cameras file
        image_folder (PATH_TYPE):
            Path to the folder of images used to create the mesh
        texture (typing.Union[PATH_TYPE, np.ndarray, None]):
            See TexturedPhotogrammetryMesh.load_texture
        render_savefolder (PATH_TYPE):
            Where to save the rendered labels
        transform_file (typing.Union[PATH_TYPE, None], optional):
            File containing the transform from local coordinates to EPSG:4978. Defaults to None.
        subset_images_savefolder (typing.Union[PATH_TYPE, None], optional):
            Where to save the subset of images for which labels are generated. Defaults to None.
        texture_column_name (typing.Union[str, None], optional):
            Column to use in vector file for texture information". Defaults to None.
        DTM_file (typing.Union[PATH_TYPE, None], optional):
            Path to a DTM file to use for ground thresholding. Defaults to None.
        ground_height_threshold (typing.Union[float, None], optional):
            Set points under this height to ground. Only applicable if DTM_file is provided. Defaults to None.
        render_ground_class (bool, optional):
            Should the ground class be included in the renders or deleted.. Defaults to False.
        textured_mesh_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the textured and subsetted mesh, if needed in the future. Defaults to None.
        ROI (typing.Union[PATH_TYPE, gpd.GeoDataFrame, shapely.MultiPolygon, None], optional):
            The region of interest to render labels for. Defaults to None.
        mesh_ROI_buffer_radius_meters (float, optional):
            The distance in meters to include around the ROI for the mesh. Defaults to 50.
        cameras_ROI_buffer_radius_meters (float, optional):
            The distance in meters to include around the ROI for the cameras. Defaults to 150.
        IDs_to_labels (typing.Union[None, dict], optional):
            Mapping between the integer labels and string values for the classes. Defaults to None.
        render_image_scale (float, optional):
            Downsample the images to this fraction of the size for increased performance but lower quality. Defaults to 1.
        mesh_downsample (float, optional):
            Downsample the mesh to this fraction of vertices for increased performance but lower quality. Defaults to 1.
        n_render_clusters (typing.Union[int, None]):
            If set, break the camera set and mesh into this many clusters before rendering. This is
            useful for large meshes that are otherwise very slow. Defaults to None.
        mesh_vis (typing.Union[PATH_TYPE, None])
            Path to save the visualized mesh instead of showing it interactively. Only applicable if vis=True. Defaults to None.
        labels_vis (typing.Union[PATH_TYPE, None])
            Defaults to None.
    """
    ## Determine the ROI
    # If the ROI is unset and the texture is a geodataframe, set the ROI to that
    if ROI is None and isinstance(texture, gpd.GeoDataFrame):
        ROI = texture
    elif ROI is None and isinstance(texture, (str, Path)):
        try:
            ROI = gpd.read_file(texture)
        except fiona.errors.DriverError:
            pass

    # If the transform filename is None, use the cameras filename instead
    # since this contains the transform information
    if transform_file is None:
        transform_file = cameras_file

    ## Create the camera set
    # This is done first because it's often faster than mesh operations which
    # makes it a good place to check for failures
    camera_set = MetashapeCameraSet(cameras_file, image_folder)

    if ROI is not None:
        # Extract cameras near the training data
        camera_set = camera_set.get_subset_ROI(
            ROI=ROI, buffer_radius=cameras_ROI_buffer_radius_meters, is_geospatial=True
        )
    # If requested, save out the images corresponding to this subset of cameras.
    # This is useful for model training.
    if subset_images_savefolder is not None:
        camera_set.save_images(subset_images_savefolder)

    # Select whether to use a class that renders by chunks or not
    MeshClass = (
        TexturedPhotogrammetryMesh
        if n_render_clusters is None
        else TexturedPhotogrammetryMeshChunked
    )

    ## Create the textured mesh
    mesh = MeshClass(
        mesh_file,
        downsample_target=mesh_downsample,
        texture=texture,
        texture_column_name=texture_column_name,
        transform_filename=transform_file,
        ROI=ROI,
        ROI_buffer_meters=mesh_ROI_buffer_radius_meters,
        IDs_to_labels=IDs_to_labels,
    )

    ## Set the ground class if applicable
    if DTM_file is not None and ground_height_threshold is not None:
        # The ground ID will be set to the next value if None, or np.nan if np.nan
        ground_ID = None if render_ground_class else np.nan
        mesh.label_ground_class(
            DTM_file=DTM_file,
            height_above_ground_threshold=ground_height_threshold,
            only_label_existing_labels=True,
            ground_class_name="GROUND",
            ground_ID=ground_ID,
            set_mesh_texture=True,
        )

    # Save the textured and subsetted mesh, if applicable
    if textured_mesh_savefile is not None:
        mesh.save_mesh(textured_mesh_savefile)

    # Show the cameras and mesh if requested
    if vis or mesh_vis_file is not None:
        mesh.vis(camera_set=camera_set, screenshot_filename=mesh_vis_file)

    # Include n_render_clusters as an optional keyword argument, if provided. This is only applicable
    # if this mesh is a TexturedPhotogrammetryMeshChunked object
    render_kwargs = (
        {} if n_render_clusters is None else {"n_clusters": n_render_clusters}
    )
    # Render the labels and save them. This is the slow step.
    mesh.save_renders(
        camera_set=camera_set,
        render_image_scale=render_image_scale,
        save_native_resolution=True,
        output_folder=render_savefolder,
        make_composites=False,
        **render_kwargs,
    )

    if vis or labels_vis_folder is not None:
        # Show some examples of the rendered labels side-by-side with the real images
        show_segmentation_labels(
            label_folder=render_savefolder,
            image_folder=image_folder,
            savefolder=labels_vis_folder,
            num_show=10,
        )


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """
    description = (
        "This script renders labels onto individual images using geospatial textures. "
        + "By default is uses the example data. All arguments are passed to "
        + "geograypher.entrypoints.workflow_functions.render_labels "
        + "which has the following documentation:\n\n"
        + render_labels.__doc__
    )
    # Ideally we'd include the defaults for each argument, but there is no help text so the
    # ArgumentDefaultsHelpFormatter formatter doesn't show them
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # Add arguments
    parser.add_argument("--mesh-file", type=Path, required=True)
    parser.add_argument("--cameras-file", type=Path, required=True)
    parser.add_argument("--image-folder", type=Path, required=True)
    parser.add_argument("--texture", type=Path, required=True)
    parser.add_argument("--render-savefolder", type=Path, required=True)
    parser.add_argument("--subset-images-savefolder", type=Path, required=True)
    parser.add_argument("--texture-column-name")
    parser.add_argument("--DTM-file")
    parser.add_argument("--ground-height-threshold", type=float, default=2.0)
    parser.add_argument("--render-ground-class", action="store_true")
    parser.add_argument("--textured-mesh-savefile")
    parser.add_argument("--ROI")
    parser.add_argument("--mesh-ROI_buffer_radius_meters", default=50, type=float)
    parser.add_argument("--cameras-ROI_buffer_radius_meters", default=100, type=float)
    parser.add_argument("--render-image-scale", type=float, default=1)
    parser.add_argument("--mesh-downsample", type=float, default=1)
    parser.add_argument("--vis", action="store_true")
    parser.add_argument("--mesh-vis-file", type=Path)
    parser.add_argument("--labels-vis-folder", type=Path)

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse command line args
    args = parse_args()
    # Pass all the arguments command line options to render_labels
    render_labels(**args.__dict__)



import os
import shutil
import tempfile
import typing
from pathlib import Path

import geopandas as gpd
import numpy as np
import rasterio as rio
from imageio import imwrite
from IPython.core.debugger import set_trace
from rasterio.features import rasterize
from rasterio.plot import reshape_as_image
from rasterio.transform import AffineTransformer
from rasterio.windows import Window
from shapely import Polygon
from tqdm import tqdm

from geograypher.constants import NULL_TEXTURE_INT_VALUE, PATH_TYPE
from geograypher.utils.files import ensure_containing_folder, ensure_folder
from geograypher.utils.io import read_image_or_numpy
from geograypher.utils.numeric import create_ramped_weighting


def create_windows(dataset_h_w, window_size, window_stride):
    windows = []
    for col_off in range(0, dataset_h_w[1], window_stride):
        for row_off in range(0, dataset_h_w[0], window_stride):
            windows.append(Window(col_off, row_off, window_size, window_size))
    return windows


def get_str_from_window(window: Window, raster_file, suffix):
    wd = window.todict()
    if suffix[0] != ".":
        suffix = "." + suffix
    window_str = f"{Path(raster_file).stem}:{wd['col_off']}:{wd['row_off']}:{wd['width']}:{wd['height']}{suffix}"
    return window_str


def parse_windows_from_files(
    files: list[Path], sep: str = ":", return_in_extent_coords: bool = True
) -> tuple[list[Window], Window]:
    """Return the boxes and extent from a list of filenames

    Args:
        files (list[Path]): List of filenames
        sep (str): Seperator between elements
        return_in_extent_coords (bool): Return in the coordinate frame of the extent

    Returns:
        tuple[list[Window], Window]: List of windows for each file and extent
    """
    # Split the coords out, currently ignorign the filename as the first element
    coords = [file.stem.split(sep)[1:] for file in files]

    # Compute the extents as the min/max of the boxes
    coords_array = np.array(coords).astype(int)

    xmin = np.min(coords_array[:, 0])
    ymin = np.min(coords_array[:, 1])
    xmax = np.max(coords_array[:, 2] + coords_array[:, 0])
    ymax = np.max(coords_array[:, 3] + coords_array[:, 1])
    extent = Window(row_off=ymin, col_off=xmin, width=xmax - xmin, height=ymax - ymin)

    if return_in_extent_coords:
        # Subtract out x and y min so it's w.r.t. the extent coordinates
        coords_array[:, 0] = coords_array[:, 0] - xmin
        coords_array[:, 1] = coords_array[:, 1] - ymin

    # Create windows from coords
    windows = [
        Window(
            col_off=coord[0],
            row_off=coord[1],
            width=coord[2],
            height=coord[3],
        )
        for coord in coords_array.astype(int)
    ]

    return windows, extent


def pad_to_full_size(img, desired_size):
    padding_size = np.array(desired_size) - np.array(img.shape[: len(desired_size)])
    if np.sum(padding_size) > 0:
        # Pad the trailing edge of the array only
        padding = [(0, width) for width in padding_size] + [(0, 0)] * (
            len(img.shape) - len(desired_size)
        )
        img = np.pad(img, padding)

    return img


def write_chips(
    raster_file,
    output_folder,
    chip_size,
    chip_stride,
    label_vector_file=None,
    label_column=None,
    label_remap=None,
    drop_transparency=True,
    remove_old=True,
    output_suffix=".JPG",
    ROI_file=None,
    background_ind=NULL_TEXTURE_INT_VALUE,
):
    if remove_old and os.path.isdir(output_folder):
        shutil.rmtree(output_folder)

    if label_vector_file is not None:
        label_gdf = gpd.read_file(label_vector_file)
    else:
        label_gdf = None

    with rio.open(raster_file, "r") as dataset:
        working_CRS = dataset.crs
        windows = create_windows(
            dataset_h_w=(dataset.height, dataset.width),
            window_size=chip_size,
            window_stride=chip_stride,
        )

        desc = f"Writing image chips to {output_folder}"
        if label_gdf is not None:
            desc = f"Writing image chips and labels to {output_folder}"
            label_gdf.to_crs(working_CRS, inplace=True)

            if label_column is not None:
                label_values = label_gdf[label_column].tolist()
            else:
                label_values = label_gdf.index.tolist()

            if label_remap is not None:
                label_values = [label_remap[old_label] for old_label in label_values]

            label_shapes = list(zip(label_gdf.geometry.values, label_values))
            labels_folder = Path(output_folder, "anns")
            output_folder = Path(output_folder, "imgs")

            ensure_folder(labels_folder)
        ensure_folder(output_folder)

        # Set up the ROI now that we have the working CRS
        if ROI_file is not None:
            ROI_gdf = gpd.read_file(ROI_file).to_crs(working_CRS)
            ROI_geometry = ROI_gdf.dissolve().geometry.values[0]
            if label_gdf is not None:
                # Crop the labels dataframe to the ROI
                label_gdf = label_gdf.intersection(ROI_geometry)
        else:
            ROI_geometry = None

        for window in tqdm(windows, desc=desc):
            if ROI_geometry is not None:
                window_transformer = AffineTransformer(dataset.window_transform(window))
                pixel_corners = (
                    (0, 0),
                    (0, chip_size),
                    (chip_size, chip_size),
                    (chip_size, 0),
                )
                geospatial_corners = [
                    window_transformer.xy(pc[0], pc[1], offset="ul")
                    for pc in pixel_corners
                ]
                geospatial_corners.append(geospatial_corners[0])
                window_polygon = Polygon(geospatial_corners)

                if not ROI_geometry.intersects(window_polygon):
                    # Skip writing this chip if it doesn't intersect the ROI
                    continue

            if label_gdf is not None:
                window_transform = dataset.window_transform(window)
                window_transformer = AffineTransformer(window_transform)
                labels_raster = rasterize(
                    label_shapes,
                    out_shape=(chip_size, chip_size),
                    transform=window_transform,
                    fill=background_ind,
                )
                labels_raster = labels_raster.astype(np.uint8)
                output_file_name = Path(
                    labels_folder,
                    get_str_from_window(
                        raster_file=raster_file, window=window, suffix=".png"
                    ),
                )
                imwrite(
                    output_file_name,
                    pad_to_full_size(labels_raster, (chip_size, chip_size)),
                )

            windowed_raster = dataset.read(window=window)
            windowed_img = reshape_as_image(windowed_raster)

            if drop_transparency and windowed_img.shape[2] == 4:
                transparency = windowed_img[..., 3]
                windowed_img = windowed_img[..., :3]
                # Set transperent regions to black
                mask = transparency == 0
                if np.all(mask):
                    continue

                windowed_img[mask, :] = 0

            output_file_name = Path(
                output_folder,
                get_str_from_window(
                    raster_file=raster_file, window=window, suffix=output_suffix
                ),
            )
            imwrite(
                output_file_name,
                pad_to_full_size(
                    windowed_img,
                    (chip_size, chip_size),
                ),
            )


def assemble_tiled_predictions(
    raster_input_file: PATH_TYPE,
    pred_files: list[PATH_TYPE],
    class_savefile: PATH_TYPE,
    num_classes: int,
    counts_savefile: typing.Union[PATH_TYPE, None] = None,
    downweight_edge_frac: float = 0.25,
    nodataval: typing.Union[int, None] = NULL_TEXTURE_INT_VALUE,
    count_dtype: type = np.uint8,
    max_overlapping_tiles: int = 4,
):
    """Take tiled predictions on disk and aggregate them into a raster

    Args:
        pred_files (list[PATH_TYPE]): List of filenames where predictions are written
        class_savefile (PATH_TYPE): Where to save the merged raster.
        counts_savefile (typing.Union[PATH_TYPE, NoneType], optional):
            Where to save the counts for the merged predictions raster.
            A tempfile will be created and then deleted if not specified. Defaults to None.
        downweight_edge_frac (float, optional): Downweight this fraction of predictions at the edge of each tile using a linear ramp. Defaults to 0.25.
        nodataval: (typing.Union[int, None]): Value for unassigned pixels. If None, will be set to len(class_names), the first unused class. Defaults to 255
        count_dtype (type, optional): What type to use for aggregation. Float uses more space but is more accurate. Defaults to np.uint8
        max_overlapping_tiles (int):
            The max number of prediction tiles that may overlap at a given point. This is used to upper bound the valud in the count matrix,
            because we use scaled np.uint8 values rather than floats for efficiency. Setting a lower value enables slightly more accuracy in the
            aggregation process, but too low can lead to overflow. Defaults to 4
    """
    # Set nodataval to the first unused class ID
    if nodataval is None:
        nodataval = num_classes

    # If the user didn't specify where to write the counts, create a tempfile that will be deleted
    if counts_savefile is None:
        # Create the containing folder if required
        ensure_containing_folder(class_savefile)
        counts_savefile_manager = tempfile.NamedTemporaryFile(
            mode="w+", suffix=".tif", dir=class_savefile.parent
        )
        counts_savefile = counts_savefile_manager.name

    # Parse the filenames to get the windows
    # TODO consider using the extent to only write a file for the minimum encolsing rectangle
    windows, extent = parse_windows_from_files(pred_files, return_in_extent_coords=True)

    # Aggregate predictions
    with rio.open(raster_input_file) as src:
        # Create file to store counts that is the same as the input raster except it has num_classes number of bands
        # TODO make this only the size of the extent computed by parse_windows_from_files
        extent_transform = src.window_transform(extent)

        with rio.open(
            counts_savefile,
            "w+",
            driver="GTiff",
            height=extent.height,
            width=extent.width,
            count=num_classes,
            dtype=count_dtype,
            crs=src.crs,
            transform=extent_transform,
        ) as dst:
            # Create
            pred_weighting_dict = {}
            for pred_file, window in tqdm(
                zip(pred_files, windows),
                desc="Aggregating raster predictions",
                total=len(pred_files),
            ):
                # Read the prediction from disk
                pred = read_image_or_numpy(pred_file)

                if pred.shape != (window.height, window.width):
                    raise ValueError("Size of pred does not match window")

                # We want to downweight portions at the edge so we create a ramped weighting mask
                # but we don't want to duplicate this computation because it's the same for each same sized chip
                if pred.shape not in pred_weighting_dict:
                    # We want to keep this as a uint8
                    pred_weighting = create_ramped_weighting(
                        pred.shape, downweight_edge_frac
                    )

                    # Allow us to get as much granularity as possible given the datatype
                    if count_dtype is not float:
                        pred_weighting = pred_weighting * (
                            np.iinfo(count_dtype).max / max_overlapping_tiles
                        )
                    # Convert weighting to desired type
                    pred_weighting_dict[pred.shape] = pred_weighting.astype(count_dtype)

                # Get weighting
                pred_weighting = pred_weighting_dict[pred.shape]

                # Update each band in the counts file within the window
                for i in range(num_classes):
                    # Bands in rasterio are 1-indexed
                    band_ind = i + 1
                    class_i_window_counts = dst.read(band_ind, window=window)
                    class_i_preds = pred == i
                    # If nothing matches this class, don't waste computation
                    if not np.any(class_i_preds):
                        continue
                    # Weight the predictions to downweight the ones at the edge
                    weighted_preds = (class_i_preds * pred_weighting).astype(
                        count_dtype
                    )
                    # Add the new predictions to the previous counts
                    class_i_window_counts += weighted_preds
                    # Write out the updated results for this window
                    dst.write(class_i_window_counts, band_ind, window=window)

    ## Convert counts file to max-class file

    with rio.open(counts_savefile, "r") as src:
        # Create a one-band file to store the index of the most predicted class
        with rio.open(
            class_savefile,
            "w",
            driver="GTiff",
            height=src.shape[0],
            width=src.shape[1],
            count=1,
            dtype=np.uint8,
            crs=src.crs,
            transform=src.transform,
            nodata=nodataval,
        ) as dst:
            # Iterate over the blocks corresponding to the tiff driver in the dataset
            # to compute the max class and write it out
            for _, window in tqdm(
                list(src.block_windows()), desc="Writing out max class"
            ):
                # Read in the counts
                counts_array = src.read(window=window)
                # Compute which pixels have no recorded predictions and mask them out
                nodata_mask = np.sum(counts_array, axis=0) == 0

                # If it's all nodata, don't write it out
                # TODO make sure this works as expected
                if np.all(nodata_mask):
                    continue

                # Compute which class had the highest counts
                max_class = np.argmax(counts_array, axis=0)
                max_class[nodata_mask] = nodataval
                # TODO, it would be good to check if it's all nodata and skip the write because that's unneeded
                dst.write(max_class, 1, window=window)



import argparse
import typing
from pathlib import Path

import numpy as np

from geograypher.cameras import MetashapeCameraSet, SegmentorPhotogrammetryCameraSet
from geograypher.constants import EXAMPLE_IDS_TO_LABELS, PATH_TYPE
from geograypher.meshes import TexturedPhotogrammetryMesh
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshChunked
from geograypher.predictors.derived_segmentors import LookUpSegmentor
from geograypher.utils.files import ensure_containing_folder
from geograypher.utils.indexing import find_argmax_nonzero_value


def aggregate_images(
    mesh_file: PATH_TYPE,
    cameras_file: PATH_TYPE,
    image_folder: PATH_TYPE,
    label_folder: PATH_TYPE,
    subset_images_folder: typing.Union[PATH_TYPE, None] = None,
    take_every_nth_camera: typing.Union[int, None] = 100,
    mesh_transform_file: typing.Union[PATH_TYPE, None] = None,
    DTM_file: typing.Union[PATH_TYPE, None] = None,
    height_above_ground_threshold: float = 2.0,
    ROI: typing.Union[PATH_TYPE, None] = None,
    ROI_buffer_radius_meters: float = 50,
    IDs_to_labels: typing.Union[dict, None] = None,
    mesh_downsample: float = 1.0,
    n_aggregation_clusters: typing.Union[int, None] = None,
    aggregate_image_scale: float = 1.0,
    aggregated_face_values_savefile: typing.Union[PATH_TYPE, None] = None,
    predicted_face_classes_savefile: typing.Union[PATH_TYPE, None] = None,
    top_down_vector_projection_savefile: typing.Union[PATH_TYPE, None] = None,
    vis: bool = False,
):
    """Aggregate labels from multiple viewpoints onto the surface of the mesh

    Args:
        mesh_file (PATH_TYPE):
            Path to the Metashape-exported mesh file
        cameras_file (PATH_TYPE):
            Path to the MetaShape-exported .xml cameras file
        image_folder (PATH_TYPE):
            Path to the folder of images used to create the mesh
        label_folder (PATH_TYPE):
            Path to the folder of labels to be aggregated onto the mesh. Must be in the same
            structure as the images
        subset_images_folder (typing.Union[PATH_TYPE, None], optional):
            Use only images from this subset. Defaults to None.
        take_every_nth_camera (typing.Union[int, None], optional):
            Downsample the camera set to only every nth camera if set. Defaults to None.
        mesh_transform_file (typing.Union[PATH_TYPE, None], optional):
            Transform from the mesh coordinates to the earth-centered, earth-fixed frame. Can be a
            4x4 matrix represented as a .csv, or a Metashape cameras file containing the
            information. Defaults to None.
        DTM_file (typing.Union[PATH_TYPE, None], optional):
            Path to a digital terrain model file to remove ground points. Defaults to None.
        height_above_ground_threshold (float, optional):
            Height in meters above the DTM to consider ground. Only used if DTM_file is set.
            Defaults to 2.0.
        ROI (typing.Union[PATH_TYPE, None], optional):
            Geofile region of interest to crop the mesh to. Defaults to None.
        ROI_buffer_radius_meters (float, optional):
            Keep points within this distance of the provided ROI object, if unset, everything will
            be kept. Defaults to 50.
        IDs_to_labels (typing.Union[dict, None], optional):
            Maps from integer IDs to human-readable class name labels. Defaults to None.
        mesh_downsample (float, optional):
            Downsample the mesh to this fraction of vertices for increased performance but lower
            quality. Defaults to 1.0.
        n_aggregation_clusters (typing.Union[int, None]):
            If set, aggregate with this many clusters. Defaults to None.
        aggregate_image_scale (float, optional):
            Downsample the labels before aggregation for faster runtime but lower quality. Defaults
            to 1.0.
        aggregated_face_values_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the aggregated image values as a numpy array. Defaults to None.
        predicted_face_classes_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the most common label per face texture as a numpy array. Defaults to None.
        top_down_vector_projection_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to export the predicted map. Defaults to None.
        vis (bool, optional):
            Show the mesh model and predicted results. Defaults to False.
    """
    ## Create the camera set
    # Do the camera operations first because they are fast and good initial error checking
    camera_set = MetashapeCameraSet(cameras_file, image_folder, validate_images=True)

    # If the ROI is not None, subset to cameras within a buffer distance of the ROI
    # TODO let get_subset_ROI accept a None ROI and return the full camera set
    if subset_images_folder is not None:
        camera_set = camera_set.get_cameras_in_folder(subset_images_folder)

    # If you only want to take every nth camera, helpful for initial testing
    if take_every_nth_camera is not None:
        camera_set = camera_set.get_subset_cameras(
            range(0, len(camera_set), take_every_nth_camera)
        )

    if ROI is not None and ROI_buffer_radius_meters is not None:
        # Extract cameras near the training data
        camera_set = camera_set.get_subset_ROI(
            ROI=ROI, buffer_radius=ROI_buffer_radius_meters
        )

    if mesh_transform_file is None:
        mesh_transform_file = cameras_file

    # Choose whether to use a mesh class that aggregates by clusters of cameras and chunks of the mesh
    MeshClass = (
        TexturedPhotogrammetryMesh
        if n_aggregation_clusters is None
        else TexturedPhotogrammetryMeshChunked
    )
    ## Create the mesh
    mesh = MeshClass(
        mesh_file,
        transform_filename=mesh_transform_file,
        ROI=ROI,
        ROI_buffer_meters=ROI_buffer_radius_meters,
        IDs_to_labels=IDs_to_labels,
        downsample_target=mesh_downsample,
    )

    # Show the mesh if requested
    if vis:
        mesh.vis(camera_set=camera_set)

    # Create a segmentor object to load in the predictions
    segmentor = LookUpSegmentor(
        base_folder=image_folder,
        lookup_folder=label_folder,
        num_classes=np.max(list(mesh.get_IDs_to_labels().keys())) + 1,
    )
    # Create a camera set that returns the segmented images instead of the original ones
    segmentor_camera_set = SegmentorPhotogrammetryCameraSet(
        camera_set, segmentor=segmentor
    )

    # Create the potentially-empty dict of kwargs to match what this class expects
    n_clusters_kwargs = (
        {} if n_aggregation_clusters is None else {"n_clusters": n_aggregation_clusters}
    )

    ## Perform aggregation, this is the slow step
    aggregated_face_labels, _ = mesh.aggregate_projected_images(
        segmentor_camera_set,
        aggregate_img_scale=aggregate_image_scale,
        **n_clusters_kwargs
    )

    # If requested, save this data
    if aggregated_face_values_savefile is not None:
        ensure_containing_folder(aggregated_face_values_savefile)
        np.save(aggregated_face_values_savefile, aggregated_face_labels)

    # Find the index of the most common class per face, with faces with no predictions set to nan
    predicted_face_classes = find_argmax_nonzero_value(
        aggregated_face_labels, keepdims=True
    )

    # If requested, label the ground faces
    if DTM_file is not None and height_above_ground_threshold is not None:
        predicted_face_classes = mesh.label_ground_class(
            labels=predicted_face_classes,
            height_above_ground_threshold=height_above_ground_threshold,
            DTM_file=DTM_file,
            ground_ID=np.nan,
            set_mesh_texture=False,
        )

    if predicted_face_classes_savefile is not None:
        ensure_containing_folder(predicted_face_classes_savefile)
        np.save(predicted_face_classes_savefile, predicted_face_classes)

    if vis:
        # Show the mesh with predicted classes
        mesh.vis(vis_scalars=predicted_face_classes)

    # TODO this should be updated to take IDs_to_labels
    mesh.export_face_labels_vector(
        face_labels=np.squeeze(predicted_face_classes),
        export_file=top_down_vector_projection_savefile,
        vis=vis,
    )


def parse_args():
    description = (
        "This script aggregates predictions from individual images onto the mesh. This aggregated "
        + "prediction can then be exported into geospatial coordinates. The default option is to "
        + "use the provided example data. All of the arguments are passed to "
        + "geograypher.entrypoints.workflow_functions.aggregate_images "
        + "which has the following documentation:\n\n"
        + aggregate_images.__doc__
    )
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description
    )
    parser.add_argument("--mesh-file", required=True)
    parser.add_argument("--cameras-file", required=True)
    parser.add_argument("--image-folder", required=True)
    parser.add_argument("--label-folder", required=True)
    parser.add_argument("--subset-images-folder")
    parser.add_argument("--take-every-nth-camera", type=int)
    parser.add_argument("--mesh-transform-file")
    parser.add_argument("--DTM-file")
    parser.add_argument("--height-above-ground-threshold", type=float, default=2)
    parser.add_argument("--ROI")
    parser.add_argument("--ROI-buffer-radius-meters", default=50, type=float)
    parser.add_argument("--IDs-to-labels", default=EXAMPLE_IDS_TO_LABELS)
    parser.add_argument("--mesh-downsample", type=float, default=1.0)
    parser.add_argument("--aggregate-image-scale", type=float, default=0.25)
    parser.add_argument("--n-aggregation-clusters", type=int)
    parser.add_argument("--aggregated-face-values-savefile", type=Path)
    parser.add_argument("--predicted-face-classes-savefile", type=Path)
    parser.add_argument(
        "--top-down-vector-projection-savefile", default="vis/predicted_map.geojson"
    )
    parser.add_argument("--vis", action="store_true")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse command line args
    args = parse_args()
    # Pass command line args to aggregate_images
    aggregate_images(**args.__dict__)



import logging
import os
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
from imageio import imread
from scipy.sparse import load_npz, save_npz

from geograypher.cameras import MetashapeCameraSet
from geograypher.cameras.segmentor import SegmentorPhotogrammetryCameraSet
from geograypher.constants import CLASS_ID_KEY, INSTANCE_ID_KEY, PATH_TYPE
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshIndexPredictions
from geograypher.predictors.derived_segmentors import TabularRectangleSegmentor
from geograypher.utils.files import ensure_containing_folder


def project_detections(
    mesh_filename: PATH_TYPE,
    cameras_filename: PATH_TYPE,
    project_to_mesh: bool = False,
    convert_to_geospatial: bool = False,
    image_folder: PATH_TYPE = None,
    detections_folder: PATH_TYPE = None,
    projections_to_mesh_filename: PATH_TYPE = None,
    projections_to_geospatial_savefilename: PATH_TYPE = None,
    default_focal_length: float = None,
    image_shape: tuple = None,
    segmentor_kwargs: dict = {},
    vis_mesh: bool = False,
    vis_geodata: bool = False,
):
    """Project per-image detections to geospatial coordinates

    Args:
        mesh_filename (PATH_TYPE):
            Path to mesh file, in local coordinates from Metashape
        cameras_filename (PATH_TYPE):
            Path to cameras file. This also contains local-to-global coordinate transform to convert
            the mesh to geospatial units.
        project_to_mesh (bool, optional):
            Execute the projection to mesh step. Defaults to False.
        convert_to_geospatial (bool, optional):
            Execute the conversion to geospatial step. Defaults to False.
        image_folder (PATH_TYPE, optional):
            Path to the folder of images used to generate the detections. TODO, see if this can be
            removed since none of this information is actually used. Defaults to None.
        detections_folder (PATH_TYPE, optional):
            Folder of detections in the DeepForest format, one per image. Defaults to None.
        projections_to_mesh_filename (PATH_TYPE, optional):
            Where to save and/or load from the data for the detections projected to the mesh faces.
            Defaults to None.
        projections_to_geospatial_savefilename (PATH_TYPE, optional):
            Where to export the geospatial detections. Defaults to None.
        default_focal_length (float, optional):
            Since the focal length is not provided in many cameras files, it can be specified.
            The units are in pixels. TODO, figure out where this information can be reliably obtained
            from. Defaults to None.
        segmentor_kwargs (dict, optional):
            Dict of keyword arguments to pass to the segmentor. Defaults to {}.
        vis_mesh (bool, optional):
            Show the mesh with detections projected onto it. Defaults to False.
        vis_geodata (bool, optional):
            Show the geospatial projection. Defaults to False.

    Raises:
        ValueError: If convert_to_geospatial but no projections to mesh are available
        FileNotFoundError: If the projections_to_mesh_filename is set and needed but not present
    """
    # Create the mesh object, which will be used for either workflow
    mesh = TexturedPhotogrammetryMeshIndexPredictions(
        mesh_filename, transform_filename=cameras_filename
    )

    # Project per-image detections to the mesh
    if project_to_mesh:
        # Create a camera set associated with the images that have detections
        camera_set = MetashapeCameraSet(
            cameras_filename,
            image_folder,
            default_sensor_params={"f": default_focal_length, "cx": 0, "cy": 0},
        )
        # Infer the image shape from the first image in the folder
        if image_shape is None:
            image_filename_list = sorted(list(Path(image_folder).glob("*.*")))
            if len(image_filename_list) > 0:
                first_file = image_filename_list[0]
                logging.info(f"loading image shape from {first_file}")
                first_image = imread(first_file)
                image_shape = first_image.shape[:2]
            else:
                raise ValueError(
                    f"No image_shape provided and folder of images {image_folder} was empty"
                )
        # Create an object that looks up the detections from a folder of CSVs or one individual one.
        # Using this, it can generate "predictions" for a given image.
        detections_predictor = TabularRectangleSegmentor(
            detection_file_or_folder=detections_folder,
            image_folder=image_folder,
            image_shape=image_shape,
            **segmentor_kwargs,
        )

        # If a file is provided for the projections, save the detection info alongside it
        if projections_to_mesh_filename is not None:
            # Export the per-image detection information as one standardized file
            detection_info_file = Path(
                projections_to_mesh_filename.parent,
                projections_to_mesh_filename.stem + "_detection_info.csv",
            )
            logging.info(f"Saving detection info to {detection_info_file}")
            detections_predictor.save_detection_data(detection_info_file)

        # Wrap the camera set so that it returns the detections rather than the original images
        detections_camera_set = SegmentorPhotogrammetryCameraSet(
            camera_set, segmentor=detections_predictor
        )
        # Project the detections to the mesh
        aggregated_prejected_images_returns = mesh.aggregate_projected_images(
            cameras=detections_camera_set, n_classes=detections_predictor.num_classes
        )
        # Get the summed (not averaged) projections
        aggregated_projections = aggregated_prejected_images_returns[1][
            "summed_projections"
        ]

        if projections_to_mesh_filename is not None:
            # Export the per-face texture to an npz file, since it's a sparse array
            ensure_containing_folder(projections_to_mesh_filename)
            save_npz(projections_to_mesh_filename, aggregated_projections)

        if vis_mesh:
            # Determine which detection is predicted for each face, if any. In cases where multiple
            # detections project to the same face, the one with the lower index will be reported
            detection_ID_per_face = np.argmax(aggregated_projections, axis=1).astype(
                float
            )
            # Mask out locations for which there are no predictions
            detection_ID_per_face[np.sum(aggregated_projections, axis=1) == 0] = np.nan
            # Show the mesh
            mesh.vis(vis_scalars=detection_ID_per_face)

    # Convert per-face projections to geospatial ones
    if convert_to_geospatial:
        # Determine if the mesh texture was computed in the last step or otherwise if it can be loaded
        if not project_to_mesh:
            if projections_to_mesh_filename is None:
                raise ValueError("No projections_to_mesh_savefilename provided")
            elif os.path.isfile(projections_to_mesh_filename):
                aggregated_projections = load_npz(projections_to_mesh_filename)
                detection_info_file = Path(
                    projections_to_mesh_filename.parent,
                    projections_to_mesh_filename.stem + "_detection_info.csv",
                )
                detection_info = pd.read_csv(detection_info_file)
            else:
                raise FileNotFoundError(
                    f"projections_to_mesh_filename {projections_to_mesh_filename} not found"
                )
        else:
            detection_info = detections_predictor.get_all_detections()

        # Convert the per-face labels to geospatial coordinates. Optionally vis and/or export
        mesh.export_face_labels_vector(
            face_labels=aggregated_projections,
            export_file=projections_to_geospatial_savefilename,
            vis=vis_geodata,
        )

        projected_geo_data = gpd.read_file(projections_to_geospatial_savefilename)
        # Merge the two dataframes so the left df's "class_ID" field aligns with the right df's
        # "instance_ID". This will add back the original data assocaited with each per-image detection
        # to the projected data.
        # Add the "_right" suffix to any of the original fields that share a name with the ones in the
        # projected data
        merged = projected_geo_data.merge(
            detection_info,
            left_on=CLASS_ID_KEY,
            right_on=INSTANCE_ID_KEY,
            suffixes=(None, "_right"),
        )
        # Drop the columns that are just an integer ID, except for "instance_ID"
        # TODO determine why "Unnamed: 0" appears
        merged.drop(
            columns=[CLASS_ID_KEY, "Unnamed: 0", "geometry_right"], inplace=True
        )

        # Save the data back out with the updated information
        merged.to_file(projections_to_geospatial_savefilename)



import argparse
import logging
from pathlib import Path

from geograypher.predictors import OrthoSegmentor


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("raster_image_file", help="Path to raster tile")
    parser.add_argument(
        "--vector-label-file",
        help="Path to vector label file. Cannot be used with --raster-label-file",
    )
    parser.add_argument(
        "--raster-label-file",
        help="Path to raster label file. Cannot be used with --vector-label-file",
    )

    parser.add_argument("--class-names", nargs="+")

    parser.add_argument(
        "--chip-size", type=int, default=2048, help="Size of chips in pixels"
    )
    parser.add_argument(
        "--training-stride-fraction",
        type=int,
        default=0.5,
        help="The stride between chips as a fraction of the tile size",
    )
    parser.add_argument(
        "--inference-stride-fraction",
        type=int,
        default=0.5,
        help="The stride between chips as a fraction of the tile size",
    )

    parser.add_argument(
        "--brightness-multiplier",
        type=float,
        default=1.0,
        help="Multiplier on chip brightness",
    )

    parser.add_argument(
        "--training-chips-folder",
        type=Path,
        help="Run chipping for training and export to this folder",
    )
    parser.add_argument(
        "--inference-chips-folder",
        type=Path,
        help="Run chipping for inference and export to this folder",
    )
    parser.add_argument(
        "--prediction-chips-folder",
        type=Path,
        help="Run aggregation using chipts from this folder",
    )
    parser.add_argument("--class-names", nargs="+")
    parser.add_argument(
        "--aggregated-savefile",
        type=Path,
        help="Save aggregated predictions to this file. Must be writable by rasterio. Only used with --prediction-chips-folder",
    )
    parser.add_argument(
        "--downweight-edge-frac",
        type=float,
        default=0.25,
        help="Downweight this fraction of predictions at the edges using a linear ramp. Only used with --prediction-chips-folder",
    )
    parser.add_argument(
        "--log-level",
        default="info",
        choices=list(logging._nameToLevel.keys()),
        help="Verbosity of printouts",
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()

    logging.basicConfig(level=args.log_level.upper())

    logging.info("Loading data for ortho segmentor class")
    ortho_seg = OrthoSegmentor(
        raster_input_file=args.raster_image_file,
        vector_label_file=args.vector_label_file,
        raster_label_file=args.raster_label_file,
        class_names=args.class_names,
        chip_size=args.chip_size,
        training_stride=int(args.training_stride_fraction * args.chip_size),
        inference_stride=int(args.inference_stride_fraction * args.chip_size),
        class_names=args.class_names,
    )

    if args.training_chips_folder is not None:
        logging.info(f"Writing training chips to {args.training_chips_folder}")
        ortho_seg.write_training_chips(
            args.training_chips_folder, brightness_multiplier=args.brightness_multiplier
        )

    if args.inference_chips_folder is not None:
        logging.info(f"Writing inference chips to {args.inference_chips_folder}")
        ortho_seg.write_inference_chips(
            args.inference_chips_folder,
            brightness_multiplier=args.brightness_multiplier,
        )

    if args.prediction_chips_folder is not None:
        logging.info(
            f"Aggregating tiled predictions from {args.prediction_chips_folder}"
            + (
                ""
                if args.aggregated_savefile is None
                else f" and saving to {args.aggregated_savefile}"
            )
        )
        ortho_seg.assemble_tiled_predictions(
            sorted(args.prediction_chips_folder.glob("*")),
            class_savefile=args.aggregated_savefile,
            downweight_edge_frac=args.downweight_edge_frac,
        )



import numpy as np
import pyvista as pv
from geopandas import GeoDataFrame
from scipy.spatial import ConvexHull
from scipy.spatial.distance import cdist
from shapely import Polygon


def create_non_overlapping_points(
    n_points, distance_thresh=1, size=10, random_seed=None
):
    np.random.seed(random_seed)
    all_points = (np.random.rand(1, 2) - 0.5) * size
    while all_points.shape[0] < n_points:
        new_point = (np.random.rand(1, 2) - 0.5) * size

        dist_from_existing = cdist(new_point, all_points)
        if np.min(dist_from_existing) > distance_thresh:
            all_points = np.concatenate((all_points, new_point), axis=0)

    return all_points


def extract_polygon(mesh: pv.PolyData):
    xy_points = mesh.points[:, :2]
    hull = ConvexHull(xy_points)
    return Polygon(xy_points[hull.vertices])


def create_scene_mesh(
    box_centers=(),
    cylinder_centers=(),
    cone_centers=(),
    cylinder_radius=0.5,
    cone_radius=0.5,
    box_size=1 / np.sqrt(2.0),
    grid_size=(20, 20),
    add_ground=True,
    ground_resolution=200,
):
    box_meshes = []
    box_polygons = []
    ID = 0.0
    for x, y in box_centers:
        x_min = x - box_size / 2.0
        x_max = x + box_size / 2.0
        y_min = y - box_size / 2.0
        y_max = y + box_size / 2.0

        box = pv.Box((x_min, x_max, y_min, y_max, 0, box_size), quads=False)
        box["ID"] = np.full(box.n_cells, fill_value=ID)
        box_meshes.append(box)
        box_polygons.append(extract_polygon(box))
        ID += 1.0

    cylinder_meshes = []
    cylinder_polygons = []
    for x, y in cylinder_centers:
        cylinder = pv.Cylinder(
            (x, y, 0.5), direction=(0, 0, 1), radius=cylinder_radius, resolution=10
        ).triangulate()
        cylinder["ID"] = np.full(cylinder.n_cells, fill_value=ID)
        cylinder_meshes.append(cylinder)
        cylinder_polygons.append(extract_polygon(cylinder))
        ID += 1.0

    cone_meshes = []
    cone_polygons = []
    for x, y in cone_centers:
        cone = pv.Cone(
            (x, y, 0.5),
            direction=(0, 0, -1),
            radius=cone_radius,
            resolution=12,
        ).triangulate()
        cone["ID"] = np.full(cone.n_cells, fill_value=ID)
        cone_meshes.append(cone)
        cone_polygons.append(extract_polygon(cone))
        ID += 1.0

    merged_mesh = pv.merge(box_meshes + cylinder_meshes + cone_meshes)
    labels_gdf = GeoDataFrame(
        {
            "name": ["cube"] * len(box_meshes)
            + ["cylinder"] * len(cylinder_meshes)
            + ["cone"] * len(cone_meshes)
        },
        geometry=box_polygons + cylinder_polygons + cone_polygons,
    )

    if add_ground:
        # Add the ground plane
        object_points = merged_mesh.points[np.isclose(merged_mesh.points[:, 2], 0)]
        grid_size_x, grid_size_y = grid_size
        # Add the corner points of the ground plane
        grid = np.meshgrid(
            np.linspace(-grid_size_x / 2, grid_size_x / 2, num=ground_resolution),
            np.linspace(-grid_size_y / 2, grid_size_y / 2, num=ground_resolution),
        )
        grid = [x.flatten() for x in grid]
        ground_points = np.vstack(grid + [np.zeros_like(grid[0])]).T

        ground_points = pv.PolyData(ground_points)
        object_points = pv.PolyData(object_points)
        combined = object_points + ground_points
        # Triangulate between all the vertices
        ground_surf = (combined).delaunay_2d()
        # Set the ID to nan
        ground_surf["ID"] = np.full(ground_surf.n_cells, fill_value=np.nan)
        # Merge the ground plane with the other actors
        merged_mesh = merged_mesh + ground_surf
    return merged_mesh, labels_gdf



import typing
import xml.etree.ElementTree as ET
from pathlib import Path

import numpy as np
import pandas as pd
import pyproj
from scipy.spatial.transform import Rotation

from geograypher.cameras import PhotogrammetryCameraSet
from geograypher.constants import EARTH_CENTERED_EARTH_FIXED_CRS, LAT_LON_CRS, PATH_TYPE
from geograypher.utils.parsing import parse_sensors, parse_transform_metashape


def update_lists(
    camera,
    image_folder,
    cam_to_world_transforms,
    image_filenames,
    sensor_IDs,
):
    transform = camera.find("transform")
    if transform is None:
        # skipping unaligned camera
        return
    # If valid, parse into numpy array
    cam_to_world_transforms.append(np.fromstring(transform.text, sep=" ").reshape(4, 4))

    # The label should contain the image path
    # TODO see if we want to do any fixup here, or punt to automate-metashape
    image_filenames.append(Path(image_folder, camera.get("label")))
    # This says which sensor model it came from
    sensor_IDs.append(int(camera.get("sensor_id")))
    # Try to get the lat lon information


class MetashapeCameraSet(PhotogrammetryCameraSet):
    def __init__(
        self,
        camera_file: PATH_TYPE,
        image_folder: PATH_TYPE,
        validate_images: bool = False,
        default_sensor_params: dict = {},
    ):
        """Parse the information about the camera intrinsics and extrinsics

        Args:
            camera_file (PATH_TYPE): Path to metashape .xml export
            image_folder: (PATH_TYPE): Path to image folder root

        Raises:
            ValueError: If camera calibration does not contain the f, cx, and cy params
        """
        # Load the xml file
        # Taken from here https://rowelldionicio.com/parsing-xml-with-python-minidom/
        tree = ET.parse(camera_file)
        root = tree.getroot()
        # first level
        chunk = root.find("chunk")
        # second level
        sensors = chunk.find("sensors")
        # Parse the sensors representation
        sensors_dict = parse_sensors(sensors, default_sensor_dict=default_sensor_params)

        # Set up the lists to populate
        image_filenames = []
        cam_to_world_transforms = []
        sensor_IDs = []

        cameras = chunk[2]
        # Iterate over metashape cameras and fill out required information
        for cam_or_group in cameras:
            if cam_or_group.tag == "group":
                for cam in cam_or_group:
                    update_lists(
                        cam,
                        image_folder,
                        cam_to_world_transforms,
                        image_filenames,
                        sensor_IDs,
                    )
            else:
                # 4x4 transform
                update_lists(
                    cam_or_group,
                    image_folder,
                    cam_to_world_transforms,
                    image_filenames,
                    sensor_IDs,
                )

        # Compute the lat lon using the transforms, because the reference values recorded in the file
        # reflect the EXIF values, not the optimized ones

        # Get the transform from the chunk to the earth-centered, earth-fixed (ECEF) frame
        chunk_to_epsg4327 = parse_transform_metashape(camera_file=camera_file)

        if chunk_to_epsg4327 is not None:
            # Compute the location of each camera in ECEF
            cam_locs_in_epsg4327 = []
            for cam_to_world_transform in cam_to_world_transforms:
                cam_loc_in_chunk = cam_to_world_transform[:, 3:]
                cam_locs_in_epsg4327.append(chunk_to_epsg4327 @ cam_loc_in_chunk)
            cam_locs_in_epsg4327 = np.concatenate(cam_locs_in_epsg4327, axis=1)[:3].T
            # Transform these points into lat-lon-alt
            transformer = pyproj.Transformer.from_crs(
                EARTH_CENTERED_EARTH_FIXED_CRS, LAT_LON_CRS
            )
            lat, lon, _ = transformer.transform(
                xx=cam_locs_in_epsg4327[:, 0],
                yy=cam_locs_in_epsg4327[:, 1],
                zz=cam_locs_in_epsg4327[:, 2],
            )
            lon_lats = list(zip(lon, lat))
        else:
            # TODO consider trying to parse from the xml
            lon_lats = None

        # Actually construct the camera objects using the base class
        super().__init__(
            cam_to_world_transforms=cam_to_world_transforms,
            intrinsic_params_per_sensor_type=sensors_dict,
            image_filenames=image_filenames,
            lon_lats=lon_lats,
            image_folder=image_folder,
            sensor_IDs=sensor_IDs,
            validate_images=validate_images,
            local_to_epsg_4978_transform=chunk_to_epsg4327,
        )


class COLMAPCameraSet(PhotogrammetryCameraSet):

    def __init__(
        self,
        cameras_file: PATH_TYPE,
        images_file: PATH_TYPE,
        image_folder: typing.Union[None, PATH_TYPE] = None,
        validate_images: bool = False,
    ):
        """
        Create a camera set from the files exported by the open-source structure-from-motion
        software COLMAP as defined here: https://colmap.github.io/format.html

        Args:
            cameras_file (PATH_TYPE):
                Path to the file containing the camera models definitions
            images_file (PATH_TYPE):
                Path to the per-image information, including the pose and which camera model is used
            image_folder (typing.Union[None, PATH_TYPE], optional):
                Path to the folder of images used to generate the reconstruction. Defaults to None.
            validate_images (bool, optional):
                Ensure that the images described in images_file are present in image_folder.
                Defaults to False.

        Raises:
            NotImplementedError: If the camera is not a Simple radial model
        """
        # Parse the csv representation of the camera models
        cameras_data = pd.read_csv(
            cameras_file,
            sep=" ",
            skiprows=[0, 1, 2],
            header=None,
            names=(
                "CAMERA_ID",
                "MODEL",
                "WIDTH",
                "HEIGHT",
                "PARAMS_F",
                "PARAMS_CX",
                "PARAMS_CY",
                "PARAMS_RADIAL",
            ),
        )
        # Parse the csv of the per-image information
        # Note that every image has first the useful information on one row and then unneeded
        # keypoint information on the following row. Therefore, the keypoints are discarded.
        images_data = pd.read_csv(
            images_file,
            sep=" ",
            skiprows=lambda x: (x in (0, 1, 2, 3) or x % 2),
            header=None,
            names=(
                "IMAGE_ID",
                "QW",
                "QX",
                "QY",
                "QZ",
                "TX",
                "TY",
                "TZ",
                "CAMERA_ID",
                "NAME",
            ),
            usecols=list(range(10)),
        )

        # TODO support more camera models
        if np.any(cameras_data["MODEL"] != "SIMPLE_RADIAL"):
            raise NotImplementedError("Not a supported camera model")

        # Parse the camera parameters, creating a dict for each distinct camera model
        sensors_dict = {}
        for _, row in cameras_data.iterrows():
            # Note that the convention in this tool is for cx, cy to be defined from the center
            # not the corner so it must be shifted
            sensor_dict = {
                "image_width": row["WIDTH"],
                "image_height": row["HEIGHT"],
                "f": row["PARAMS_F"],
                "cx": row["PARAMS_CX"] - row["WIDTH"] / 2,
                "cy": row["PARAMS_CY"] - row["HEIGHT"] / 2,
                "distortion_params": {"r": row["PARAMS_RADIAL"]},
            }
            sensors_dict[row["CAMERA_ID"]] = sensor_dict

        # Parse the per-image information
        cam_to_world_transforms = []
        sensor_IDs = []
        image_filenames = []

        for _, row in images_data.iterrows():
            # Convert from the quaternion representation to the matrix one. Note that the W element
            # is the first one in the COLMAP convention but the last one in scipy.
            rot_mat = Rotation.from_quat(
                (row["QX"], row["QY"], row["QZ"], row["QW"])
            ).as_matrix()
            # Get the camera translation
            translation_vec = np.array([row["TX"], row["TY"], row["TZ"]])

            # Create a 4x4 homogenous matrix representing the world_to_cam transform
            world_to_cam = np.eye(4)
            # Populate the sub-elements
            world_to_cam[:3, :3] = rot_mat
            world_to_cam[:3, 3] = translation_vec
            # We need the cam to world transform. Since we're using a 4x4 representation, we can
            # just invert the matrix
            cam_to_world = np.linalg.inv(world_to_cam)
            cam_to_world_transforms.append(cam_to_world)

            # Record which camera model is used and the image filename
            sensor_IDs.append(row["CAMERA_ID"])
            image_filenames.append(Path(image_folder, row["NAME"]))

        # Instantiate the camera set
        super().__init__(
            cam_to_world_transforms=cam_to_world_transforms,
            intrinsic_params_per_sensor_type=sensors_dict,
            image_filenames=image_filenames,
            sensor_IDs=sensor_IDs,
            image_folder=image_folder,
            validate_images=validate_images,
        )



import xml.etree.ElementTree as ET
from copy import copy

import numpy as np


def make_4x4_transform(rotation_str: str, translation_str: str, scale_str: str = "1"):
    """Convenience function to make a 4x4 matrix from the string format used by Metashape

    Args:
        rotation_str (str): Row major with 9 entries
        translation_str (str): 3 entries
        scale_str (str, optional): single value. Defaults to "1".

    Returns:
        np.ndarray: (4, 4) A homogenous transform mapping from cam to world
    """
    rotation_np = np.fromstring(rotation_str, sep=" ")
    rotation_np = np.reshape(rotation_np, (3, 3))

    if not np.isclose(np.linalg.det(rotation_np), 1.0, atol=1e-8, rtol=0):
        raise ValueError(
            f"Inproper rotation matrix with determinant {np.linalg.det(rotation_np)}"
        )

    translation_np = np.fromstring(translation_str, sep=" ")
    scale = float(scale_str)
    transform = np.eye(4)
    transform[:3, :3] = rotation_np * scale
    transform[:3, 3] = translation_np
    return transform


def parse_transform_metashape(camera_file):
    tree = ET.parse(camera_file)
    root = tree.getroot()
    # first level
    components = root.find("chunk").find("components")

    assert len(components) == 1
    transform = components.find("component").find("transform")
    if transform is None:
        return None

    rotation = transform.find("rotation").text
    translation = transform.find("translation").text
    scale = transform.find("scale").text

    local_to_epgs_4978_transform = make_4x4_transform(rotation, translation, scale)

    return local_to_epgs_4978_transform


def parse_sensors(sensors, default_sensor_dict=None):
    sensors_dict = {}

    for sensor in sensors:
        sensor_dict = {}

        sensor_dict["image_width"] = int(sensor[0].get("width"))
        sensor_dict["image_height"] = int(sensor[0].get("height"))
        calibration = sensor.find("calibration")

        if calibration is None:
            if default_sensor_dict is not None:
                for k, v in default_sensor_dict.items():
                    sensor_dict[k] = v
            else:
                sensor_dict = None
        else:
            sensor_dict["f"] = float(calibration.find("f").text)

            # Extract the objects for the principal points
            cx = calibration.find("cx")
            cy = calibration.find("cx")

            try:
                # If they are set, use the value. Otherwise, use the default.
                # TODO handle the case where no default is provided explicitly
                sensor_dict["cx"] = (
                    float(cx.text) if cx is not None else default_sensor_dict["cx"]
                )
                sensor_dict["cy"] = (
                    float(cy.text) if cy is not None else default_sensor_dict["cy"]
                )
                # Get potentially-empty dict of distortion parameters
                sensor_dict["distortion_params"] = {
                    calibration[i].tag: float(calibration[i].text)
                    for i in range(3, len(calibration))
                }
            except KeyError:
                sensor_dict = None

        sensor_ID = int(sensor.get("id"))
        sensors_dict[sensor_ID] = sensor_dict
    return sensors_dict



import os
import typing
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from imageio import imread
from PIL import Image
from skimage.transform import resize

from geograypher.constants import PATH_TYPE
from geograypher.predictors import Segmentor
from geograypher.utils.files import ensure_containing_folder


class BrightnessSegmentor(Segmentor):
    def __init__(self, brightness_threshold: float = np.sqrt(0.75)):
        self.brightness_threshold = brightness_threshold
        self.num_classes = 2

    def segment_image(self, image: np.ndarray, **kwargs):
        image_brightness = np.linalg.norm(image, axis=-1)
        thresholded_image = image_brightness > self.brightness_threshold
        class_index_image = thresholded_image.astype(np.uint8)
        one_hot_image = self.inds_to_one_hot(class_index_image)
        return one_hot_image


class LookUpSegmentor(Segmentor):
    def __init__(self, base_folder, lookup_folder, num_classes=10):
        self.base_folder = Path(base_folder)
        self.lookup_folder = lookup_folder
        self.num_classes = num_classes

    def segment_image(self, image: np.ndarray, filename: PATH_TYPE, image_scale: float):
        relative_path = Path(filename).relative_to(self.base_folder)
        lookup_path = Path(self.lookup_folder, relative_path)
        lookup_path = lookup_path.with_suffix(".png")

        image = imread(lookup_path)
        if image_scale != 1:
            image = resize(
                image,
                (int(image.shape[0] * image_scale), int(image.shape[1] * image_scale)),
                order=0,  # Nearest neighbor interpolation
            )
        one_hot_image = self.inds_to_one_hot(image, num_classes=self.num_classes)
        return one_hot_image


class ImageIDSegmentor(Segmentor):
    """Setmentor that returns an image full of the index of that camera in the set"""

    def __init__(self, image_filenames: typing.List[PATH_TYPE]):
        """Set up the segmentor based on a list of image filenames.

        Args:
            image_filenames (typing.List[PATH_TYPE]):
                The list of absolute image paths. In the segmentation stage, the returned value with
                be the index of an image within this list.
        """
        self.image_filenames = image_filenames

    def segment_image(self, image: np.ndarray, filename: PATH_TYPE, image_scale: float):
        # Get the shape of the image without reading it into memory
        img_handler = Image.open(filename)
        w, h = img_handler.size

        # Get the index of the image in the list
        image_index = self.image_filenames.index(filename)
        if image_index is None:
            raise ValueError(f"Image {filename} not found in list")

        # Get the scaled shape to output
        output_shape = (int(h * image_scale), int(w * image_scale))
        # Create an array of the appropriate size filled with the value of the index within the list
        ID_image = np.full(output_shape, fill_value=image_index, dtype=int)
        return ID_image


class TabularRectangleSegmentor(Segmentor):
    def __init__(
        self,
        detection_file_or_folder: PATH_TYPE,
        image_shape: tuple,
        label_key: str = "instance_ID",
        image_path_key: str = "image_path",
        imin_key: str = "ymin",
        imax_key: str = "ymax",
        jmin_key: str = "xmin",
        jmax_key: str = "xmax",
        detection_file_extension: str = "csv",
        strip_image_extension: bool = False,
        use_absolute_filepaths: bool = False,
        split_bbox: bool = True,
        image_folder: typing.Union[PATH_TYPE, None] = None,
    ):
        """Lookup rectangular bounding boxes corresponding to detections from a CSV or folder of them.

        Args:
            detection_file_or_folder (PATH_TYPE):
                Path to the CSV file with detections or a folder thereof
            image_shape (tuple):
                The (height, width) shape of the image in pixels.
            label_key (str, optional):
                The column that corresponds to the class. Defaults to "label".
            image_path_key (str, optional):
                The column that has the image filename. Defaults to "image_path".
            imin_key (str, optional):
                Column of the minimum i dimension. Defaults to "ymin".
            imax_key (str, optional):
                Column of the max i dimension. Defaults to "ymax".
            jmin_key (str, optional):
                Column of the min j dimension. Defaults to "xmin".
            jmax_key (str, optional):
                Column of the max j dimension. Defaults to "xmax".
            detection_file_extension (str, optional):
                File extension of the detection files. Defaults to "csv".
            strip_image_extension (bool, optional):
                Remove the extension from the image filenames. Defaults to True.
            use_absolute_filepaths (bool, optional):
                Add the absolute path from the image folder to the filenames. Defaults to False.
            split_bbox (bool, optional):
                Split the bounding box from one column rather than having seperate columns for imin,
                imax, jmin, jmax. Defaults to True.
            image_folder (PATH_TYPE, optional): Path to the image folder. Defaults to None.
        """
        self.image_shape = image_shape

        self.label_key = label_key
        self.image_path_key = image_path_key
        self.imin_key = imin_key
        self.imax_key = imax_key
        self.jmin_key = jmin_key
        self.jmax_key = jmax_key
        self.split_bbox = split_bbox

        # Load the detections
        self.labels_df = self.load_detection_files(
            detection_file_or_folder=detection_file_or_folder,
            detection_file_extension=detection_file_extension,
            image_folder=image_folder,
            use_absolute_filepaths=use_absolute_filepaths,
            strip_image_extension=strip_image_extension,
            image_path_key=image_path_key,
        )

        # Group the predictions
        self.grouped_labels_df = self.labels_df.groupby(by=self.image_path_key)

        # List the images
        self.image_names = list(self.grouped_labels_df.groups.keys())
        # Record the class names and number of classes
        self.class_names = np.unique(self.labels_df[self.label_key]).tolist()
        self.num_classes = len(self.class_names)

    def load_detection_files(
        self,
        detection_file_or_folder: PATH_TYPE,
        detection_file_extension: str,
        image_folder: PATH_TYPE,
        use_absolute_filepaths: bool,
        strip_image_extension: bool,
        image_path_key: str,
    ):
        # Determine whether the input is a file or folder
        if Path(detection_file_or_folder).is_file():
            # If it's a file, make a one-length list
            files = [detection_file_or_folder]
        else:
            # List all the files in the folder with the requested extesion
            files = sorted(
                Path(detection_file_or_folder).glob("*" + detection_file_extension)
            )

        # Read the individual files
        dfs = [pd.read_csv(f) for f in files]

        # Concatenate the dataframes into one
        labels_df = pd.concat(dfs, ignore_index=True)

        # Add an sequential instance ID column if not present
        if "instance_ID" not in labels_df.columns:
            labels_df["instance_ID"] = labels_df.index

        # Prepend the image folder to the image filenames if requested to make an absolute filepath
        if image_folder is not None and use_absolute_filepaths:
            absolute_filepaths = [
                str(Path(image_folder, img_path))
                for img_path in labels_df[image_path_key].tolist()
            ]
            labels_df[image_path_key] = absolute_filepaths

        # Strip the extension from the image filenames if requested
        if strip_image_extension:
            image_path_without_ext = [
                str(Path(img_path).with_suffix(""))
                for img_path in labels_df[image_path_key].tolist()
            ]
            labels_df[image_path_key] = image_path_without_ext

        return labels_df

    def get_all_detections(self) -> pd.DataFrame:
        """Return the aggregated detections dataframe"""
        return self.labels_df

    def save_detection_data(self, output_csv_file: PATH_TYPE):
        """Save the aggregated detections to a file

        Args:
            output_csv_file (PATH_TYPE):
                A path to a CSV file to save the detections to. The containing folder will be
                created if needed.
        """
        ensure_containing_folder(output_csv_file)
        self.labels_df.to_csv(output_csv_file)

    def get_corners(self, data, as_int=True):
        if self.split_bbox:
            # TODO split row
            bbox = data["bbox"]
            bbox = bbox[1:-1]
            splits = bbox.split(", ")
            jmin, imin, width, height = [float(s) for s in splits]

            imax = imin + height
            jmax = jmin + width

            imin = imin
            jmin = jmin
        else:
            imin = data[self.imin_key]
            imax = data[self.imax_key]
            jmin = data[self.jmin_key]
            jmax = data[self.jmax_key]

        corners = imin, jmin, imax, jmax
        if as_int:
            corners = list(map(int, corners))

        return corners

    def segment_image(self, image, filename, image_scale, vis=False):
        output_shape = self.image_shape
        label_image = np.full(output_shape, fill_value=np.nan, dtype=float)

        name = filename.name
        if name in self.image_names:
            df = self.grouped_labels_df.get_group(name)
        # Return an all-zero segmentation image
        else:
            return label_image

        for _, row in df.iterrows():
            label = row[self.label_key]
            label_ind = self.class_names.index(label)

            imin, jmin, imax, jmax = self.get_corners(
                row,
            )

            label_image[imin:imax, jmin:jmax] = label_ind

        if vis:
            plt.imshow(label_image, vmin=0, vmax=10, cmap="tab10")
            plt.show()

        if image_scale != 1.0:
            output_size = (int(image_scale * x) for x in label_image.shape[:2])
            label_image = resize(label_image, output_size, order=0)

        return label_image

    def get_detection_centers(self, filename):
        """_summary_

        Args:
            filename (_type_): _description_

        Returns:
            _type_: (n,2) array for (i,j) centers for each detection
        """
        if filename not in self.image_names:
            # Empty array of detection centers
            return np.zeros((0, 2))

        # Extract the corresponding dataframe
        df = self.grouped_labels_df.get_group(filename)

        all_corners = []
        for _, row in df.iterrows():
            corners = self.get_corners(row, as_int=False)
            all_corners.append(corners)

        all_corners = zip(*all_corners)
        all_corners = [np.array(x) for x in all_corners]

        imin, jmin, imax, jmax = all_corners

        # Average the left-right, top-bottom pairs
        centers = np.vstack([(imin + imax) / 2, (jmin + jmax) / 2]).T
        return centers



import typing

import numpy as np
import pyvista as pv


def create_ramped_weighting(
    rectangle_shape: typing.Tuple[int, int], ramp_dist_frac: float
) -> np.ndarray:
    """Create a ramped weighting that is higher toward the center with a max value of 1 at a fraction from the edge

    Args:
        rectangle_shape (typing.Tuple[int, int]): Size of rectangle to create a mask for
        ramp_dist_frac (float): Portions at least this far from an edge will have full weight

    Returns:
        np.ndarray: An array representing the weighting from 0-1
    """
    i_ramp = np.clip(np.linspace(0, 1 / ramp_dist_frac, num=rectangle_shape[0]), 0, 1)
    j_ramp = np.clip(np.linspace(0, 1 / ramp_dist_frac, num=rectangle_shape[1]), 0, 1)

    i_ramp = np.minimum(i_ramp, np.flip(i_ramp))
    j_ramp = np.minimum(j_ramp, np.flip(j_ramp))

    i_ramp = np.expand_dims(i_ramp, 1)
    j_ramp = np.expand_dims(j_ramp, 0)

    ramped_weighting = np.minimum(i_ramp, j_ramp)
    return ramped_weighting


def compute_approximate_ray_intersection(
    A: np.array, a: np.array, B: np.array, b: np.array, vis=False
):
    # https://palitri.com/vault/stuff/maths/Rays%18closest%20point.pdf
    c = B - A

    aa = np.dot(a, a)
    ab = np.dot(a, b)
    ac = np.dot(a, c)
    bb = np.dot(b, b)
    bc = np.dot(b, c)

    denominator = aa * bb - ab * ab

    a_scaler = (-ab * bc + ac * bb) / denominator
    b_scaler = (ab * ac - bc * aa) / denominator

    D = A + a_scaler * a
    E = B + b_scaler * b

    dist = np.linalg.norm(D - E)
    # Check that it's in front of the camera
    # TODO I'm not sure if this will handle co-linear cases, probably should check for that
    valid = a_scaler > -1 and b_scaler > 0

    if vis:
        points = np.vstack([A, D, B, E, D, E])

        plotter = pv.Plotter()
        plotter.add_lines(points)
        plotter.add_points(points)
        plotter.background_color = "black"
        plotter.show()

    return dist, valid


def triangulate_rays_lstsq(starts, directions):
    # https://github.com/tensorflow/graphics/blob/master/tensorflow_graphics/geometry/representation/ray.py#L284-L371

    # Ensure that we have unit vectors specifying the direction of the ray

    # Build a cross product matrix for each direction, this is the left side
    # The right side is the dot product between this and the start ray

    As = []
    bs = []
    # TODO build this all at once rather than iteratively
    for start, direction in zip(starts, directions):
        x, y, z = direction
        # https://math.stackexchange.com/questions/3764426/matrix-vector-multiplication-cross-product-problem
        cross_matrix = np.array(
            [
                [0, -z, y],
                [z, 0, -x],
                [-y, x, 0],
            ]
        )
        As.append(cross_matrix)
        b = cross_matrix @ start
        bs.append(b)

    A = np.concatenate(As, axis=0)
    b = np.concatenate(bs, axis=0)

    x, _, _, _ = np.linalg.lstsq(A, b)
    return x


def compute_3D_triangle_area_vectorized(corners: np.ndarray, return_z_proj_area=True):
    """_summary_

    Args:
        corners (np.ndarray): (n_faces, n)
        return_z_proj_area (bool, optional): _description_. Defaults to True.

    Returns:
        _type_: _description_
    """
    A, B, C = corners
    # https://math.stackexchange.com/questions/2152754/calculate-3d-triangle-area-by-determinant
    u = B - A
    v = C - A

    # Save for future computation
    u0v1_min_u1v0 = u[0] * v[1] - u[1] * v[0]
    area = (
        1
        / 2
        * np.sqrt(
            np.power(u[1] * v[2] - u[2] * v[1], 2)
            + np.power(u[2] * v[0] - u[0] * v[2], 2)
            + np.power(u0v1_min_u1v0, 2)
        )
    )

    if return_z_proj_area:
        area_z_proj = np.abs(u0v1_min_u1v0) / 2
        return area, area_z_proj

    return area


def compute_3D_triangle_area(corners, return_z_proj_area=True):
    A, B, C = corners
    # https://math.stackexchange.com/questions/2152754/calculate-3d-triangle-area-by-determinant
    u = B - A
    v = C - A

    # Save for future computation
    u0v1_min_u1v0 = u[0] * v[1] - u[1] * v[0]
    area = (
        1
        / 2
        * np.sqrt(
            np.power(u[1] * v[2] - u[2] * v[1], 2)
            + np.power(u[2] * v[0] - u[0] * v[2], 2)
            + np.power(u0v1_min_u1v0, 2)
        )
    )

    if return_z_proj_area:
        area_z_proj = np.abs(u0v1_min_u1v0) / 2
        return area, area_z_proj

    return area



import logging
import typing

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pyproj
import rasterio as rio
from geopandas import GeoDataFrame, GeoSeries
from IPython.core.debugger import set_trace
from rasterstats import zonal_stats
from shapely import (
    Geometry,
    MultiPolygon,
    Polygon,
    difference,
    intersection,
    make_valid,
    union,
)
from shapely.geometry import box
from tqdm import tqdm

from geograypher.constants import LAT_LON_CRS, PATH_TYPE


def ensure_projected_CRS(geodata: gpd.GeoDataFrame):
    """Returns a projected geodataframe from the provided geodataframe by converting it to
    ESPG:4326 (if not already) and determining the projected CRS from the point
    coordinates.

    Args:
        geodata (gpd.GeoDataGrame): Original geodataframe that is potentially unprojected
    Returns:
        gpd.GeoDataGrame: projected geodataframe
    """
    # If CRS is projected return immediately
    if geodata.crs.is_projected:
        return geodata

    # If CRS is geographic and not long-lat, convert it to long-lat
    if geodata.crs.is_geographic and geodata.crs != LAT_LON_CRS:
        geodata = geodata.to_crs(LAT_LON_CRS)

    # Convert geographic long-lat CRS to projected CRS
    point = geodata["geometry"][0].centroid
    geometric_crs = get_projected_CRS(lon=point.x, lat=point.y)
    return geodata.to_crs(geometric_crs)


def get_projected_CRS(lat, lon, assume_western_hem=True):
    if assume_western_hem and lon > 0:
        lon = -lon
    # https://gis.stackexchange.com/questions/190198/how-to-get-appropriate-crs-for-a-position-specified-in-lat-lon-coordinates
    epgs_code = 32700 - round((45 + lat) / 90) * 100 + round((183 + lon) / 6)
    crs = pyproj.CRS.from_epsg(epgs_code)
    return crs


def convert_CRS_3D_points(points, input_CRS, output_CRS):
    transformer = pyproj.Transformer.from_crs(input_CRS, output_CRS)

    # Transform the coordinates
    points_in_output_CRS = transformer.transform(
        xx=points[:, 0],
        yy=points[:, 1],
        zz=points[:, 2],
    )
    # Stack and transpose
    points_in_output_CRS = np.vstack(points_in_output_CRS).T
    return points_in_output_CRS


def ensure_non_overlapping_polygons(
    geometries: typing.Union[typing.List[Geometry], gpd.GeoDataFrame],
    inplace: bool = False,
):
    # Make sure geometries is a list of shapely objects
    if isinstance(geometries, gpd.GeoDataFrame):
        original_gdf = geometries
        geometries = geometries.geometry.tolist()
    else:
        original_gdf = None

    output_geometries = [None] * len(geometries)
    union_of_added_geoms = MultiPolygon()

    areas = [geom.area for geom in geometries]
    sorted_inds = np.argsort(areas)

    for ind in sorted_inds:
        # Get the input geometry and ensure it's valid
        geom = make_valid(geometries[ind])
        # Subtract the union of all
        geom_to_add = difference(geom, union_of_added_geoms)
        output_geometries[ind] = geom_to_add
        # Add the original geom, not the difference'd one, to avoid boundary artifacts
        union_of_added_geoms = union(geom, union_of_added_geoms)

    if original_gdf is None:
        return output_geometries
    elif inplace:
        original_gdf.geometry = output_geometries
    else:
        output_gdf = original_gdf.copy()
        output_gdf.geometry = output_geometries
        return output_gdf


def find_union_of_intersections(list_of_multipolygons, crs, vis=False):
    all_intersections = MultiPolygon()
    for i, multipolygon_a in enumerate(list_of_multipolygons):
        for multipolygon_b in list_of_multipolygons[:i]:
            new_intersection = intersection(multipolygon_a, multipolygon_b)
            all_intersections = union(all_intersections, new_intersection)
    if vis:
        geopandas_all_intersections = GeoDataFrame(
            geometry=[all_intersections], crs=crs
        )
        geopandas_all_intersections.plot()
        plt.show()
    return all_intersections


def intersects_union_of_polygons(
    query_polygons: GeoDataFrame,
    region_polygon: typing.Union[GeoDataFrame, GeoSeries, Polygon, MultiPolygon],
):
    if isinstance(region_polygon, GeoDataFrame):
        region_polygon.plot()
        # Try to make geometries valid
        region_polygon.geometry = region_polygon.buffer(0)
        region_polygon = region_polygon.dissolve()
        region_polygon = region_polygon.geometry[0]

    # Find the polygons that are within the bounds of the raster
    intersection = query_polygons.intersection(region_polygon)
    empty_geometry = intersection.is_empty.to_numpy()
    within_bounds_IDs = np.where(np.logical_not(empty_geometry))[0]
    return within_bounds_IDs


def coerce_to_geoframe(potential_geoframe):
    # Try to load the vector data if it's not a geodataframe
    if not isinstance(potential_geoframe, GeoDataFrame):
        potential_geoframe = gpd.read_file(potential_geoframe)
    return potential_geoframe


def get_overlap_raster(
    unlabeled_df: typing.Union[PATH_TYPE, GeoDataFrame],
    classes_raster: PATH_TYPE,
    num_classes: typing.Union[None, int] = None,
    normalize: bool = False,
) -> (np.ndarray, np.ndarray):
    """Get the overlap for each polygon in the unlabeled DF with each class in the raster

    Args:
        unlabeled_df (typing.Union[PATH_TYPE, GeoDataFrame]):
            Dataframe or path to dataframe containing geometries per object
        classes_raster (PATH_TYPE): Path to a categorical raster
        num_classes (typing.Union[None, int], optional):
            Number of classes, if None defaults to the highest overlapping class. Defaults to None.
        normalize (bool, optional): Normalize counts matrix from pixels to fraction. Defaults to False.

    Returns:
        np.ndarray: (n_valid, n_classes) counts per polygon per class
        np.ndarray: (n_valid,) indices into the original array for polygons with non-null predictions
    """
    unlabeled_df = coerce_to_geoframe(unlabeled_df)

    with rio.open(classes_raster, "r") as src:
        raster_crs = src.crs
        # Note this is a shapely object with no CRS, but we ensure
        # the vectors are converted to the same CRS
        raster_bounds = box(*src.bounds)

    # Ensure that the vector data is in the same CRS as the raster
    if raster_crs != unlabeled_df.crs:
        # Avoid doing this in place because we don't want to modify the input dataframe
        # This should properly create a copy
        unlabeled_df = unlabeled_df.to_crs(raster_crs)

    # Compute which polygons intersect the raster region
    within_bounds_IDs = intersects_union_of_polygons(unlabeled_df, raster_bounds)

    # Compute the stats
    stats = zonal_stats(
        unlabeled_df.iloc[within_bounds_IDs], str(classes_raster), categorical=True
    )

    # Find which polygons have non-null class predictions
    # Due to nondata regions, some polygons within the region may not have class information
    valid_prediction_IDs = np.where([x != {} for x in stats])[0]

    # Determine the number of classes if not set
    if num_classes is None:
        # Find the max value that show up in valid predictions
        num_classes = 1 + np.max(
            [np.max(list(stats[i].keys())) for i in valid_prediction_IDs]
        )

    # Build the counts matrix for non-null predictions
    counts_matrix = np.zeros((len(valid_prediction_IDs), num_classes))

    # Fill the counts matrix
    for i in valid_prediction_IDs:
        for j, count in stats[i].items():
            counts_matrix[i, j] = count

    # Bookkeeping to find the IDs that were both within the raster and non-null
    valid_IDs_in_original = within_bounds_IDs[valid_prediction_IDs]

    if normalize:
        counts_matrix = counts_matrix / np.sum(counts_matrix, axis=1, keepdims=True)

    return counts_matrix, valid_IDs_in_original


# https://gis.stackexchange.com/questions/421888/getting-the-percentage-of-how-much-areas-intersects-with-another-using-geopandas
def get_overlap_vector(
    unlabeled_df: GeoDataFrame,
    classes_df: GeoDataFrame,
    class_column: str,
    normalize: bool = False,
) -> (np.ndarray, np.ndarray):
    """
    For each element in unlabeled df, return the fractional overlap with each class in
    classes_df


    Args:
        unlabeled_df (GeoDataFrame): A dataframe of geometries
        classes_df (GeoDataFrame): A dataframe of classes
        class_column (str, optional): Which column in the classes_df to use. Defaults to "names".
        normalize (bool, optional): Normalize counts matrix from area to fraction. Defaults to False.

    Returns:
        np.ndarray: (n_valid, n_classes) counts per polygon per class
        np.ndarray: (n_valid,) indices into the original array for polygons with non-null predictions
    """

    ## Preprocessing
    # Ensure that both a geodataframes
    unlabeled_df = coerce_to_geoframe(unlabeled_df)
    classes_df = coerce_to_geoframe(classes_df)

    unlabeled_df = ensure_projected_CRS(unlabeled_df)
    if classes_df.crs != unlabeled_df.crs:
        classes_df = classes_df.to_crs(unlabeled_df.crs)

    unlabeled_df.geometry = unlabeled_df.geometry.simplify(0.01)
    classes_df.geometry = classes_df.geometry.simplify(0.01)

    if class_column not in classes_df.columns:
        raise ValueError(f"Class column `{class_column}` not in {classes_df.columns}")

    logging.info(
        "Computing the intersection of the unlabeled polygons with the labeled region"
    )
    # Find which unlabeled polygons intersect with the labeled region
    intersection_IDs = intersects_union_of_polygons(unlabeled_df, classes_df)
    logging.info("Finished computing intersection")
    # Extract only these polygons
    unlabeled_df_intersecting_classes = unlabeled_df.iloc[intersection_IDs]
    unlabeled_df_intersecting_classes["index"] = unlabeled_df_intersecting_classes.index

    # Add area field to each
    unlabeled_df_intersecting_classes["unlabeled_area"] = (
        unlabeled_df_intersecting_classes.area
    )

    # Find the intersecting geometries
    # We want only the ones that have some overlap with the unlabeled geometry, but I don't think that can be specified
    logging.info("computing overlay")
    overlay = gpd.overlay(
        classes_df,
        unlabeled_df_intersecting_classes,
        how="union",
        keep_geom_type=False,
    )
    # Drop the rows that only contain information from the class_labels
    overlay = overlay.dropna(subset="index")

    # TODO look more into this part, something seems wrong
    overlay["overlapping_area"] = overlay.area
    # overlay["per_class_area_fraction"] = (
    #    overlay["overlapping_area"] / overlay["unlabeled_area"]
    # )
    # Aggregating the results
    # WARNING Make sure that this is a list and not a tuple or it gets considered one key
    logging.info("computing groupby")

    # If the two dataframes have a column with the same name, they will be renamed
    if (
        f"{class_column}_1" in overlay.columns
        and f"{class_column}_2" in overlay.columns
    ):
        aggregatation_class_column = f"{class_column}_1"
    else:
        aggregatation_class_column = class_column

    # Groupby and aggregate
    grouped_by = overlay.groupby(by=["index", aggregatation_class_column])
    aggregated = grouped_by.agg({"overlapping_area": "sum"})

    # Extract the original class names
    unique_class_names = sorted(classes_df[class_column].unique().tolist())
    # And the indices from the original dataframe. This is relavent if the input
    # dataframe was a subset of an original one
    unique_index_values = sorted(unlabeled_df_intersecting_classes.index.tolist())

    counts_matrix = np.zeros(
        (len(unlabeled_df_intersecting_classes), len(unique_class_names))
    )

    for r in aggregated.iterrows():
        (index, class_name), area = r
        # The index is the index from the original unlabeled dataset, but we need the index into the subset
        unlabled_object_ind = unique_index_values.index(index)
        # Transform the class name into a class index
        class_ind = unique_class_names.index(class_name)
        # Set the value to the area of the overlap between that unlabeled object and given class
        counts_matrix[unlabled_object_ind, class_ind] = float(area.iloc[0])

    if normalize:
        counts_matrix = counts_matrix / np.sum(counts_matrix, axis=1, keepdims=True)

    return counts_matrix, intersection_IDs, unique_class_names


# https://stackoverflow.com/questions/60288953/how-to-change-the-crs-of-a-raster-with-rasterio
def reproject_raster(in_path, out_path, out_crs=pyproj.CRS.from_epsg(4326)):
    """ """
    logging.warn("Starting to reproject raster")
    # reproject raster to project crs
    with rio.open(in_path) as src:
        src_crs = src.crs
        transform, width, height = rio.warp.calculate_default_transform(
            src_crs, out_crs, src.width, src.height, *src.bounds
        )
        kwargs = src.meta.copy()

        kwargs.update(
            {"crs": out_crs, "transform": transform, "width": width, "height": height}
        )

        with rio.open(out_path, "w", **kwargs) as dst:
            for i in tqdm(range(1, src.count + 1), desc="Reprojecting bands"):
                rio.warp.reproject(
                    source=rio.band(src, i),
                    destination=rio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=out_crs,
                    resampling=rio.warp.Resampling.nearest,
                )
    logging.warn("Done reprojecting raster")



import logging
import typing

import numpy as np
import shapely
from shapely import simplify, unary_union
from tqdm import tqdm

logger = logging.getLogger("geometric")


def batched_unary_union(
    geometries: typing.List[shapely.Geometry],
    batch_size: int,
    grid_size: typing.Union[None, float] = None,
    subsequent_batch_size: int = 4,
    sort_by_loc: bool = False,
    simplify_tol: float = 0,
    verbose: bool = False,
) -> shapely.MultiPolygon:
    """Roughly replicate the functionality of shapely.unary_union using a batched implementation

    Args:
        geometries (typing.List[shapely.Geometry]): Geometries to aggregate
        batch_size (int): The batch size for the first aggregation
        grid_size (typing.Union[None, float]): grid size passed to unary_union
        subsequent_batch_size (int, optional): The batch size for subsequent (recursive) batches. Defaults to 4.
        sort_by_loc (bool, optional): Should the polygons be sorted by location to have a higher likelihood of merging. Defaults to False.
        simplify_tol (float, optional): How much to simplify in intermediate steps
        verbose (bool, optional): Should additional print outs be provided

    Returns:
        shapely.MultiPolygon: The merged multipolygon
    """

    # If the geoemtry is already one entry or empty
    if len(geometries) <= 1:
        # Run unary_union to ensure that edge cases such as empty geometries are handled in the expected way
        return unary_union(geometries)

    # Sort the polygons by their least x coordinates (any bound could be used)
    # The goal of this is to have a higher likelihood of grouping objects together and removing interior coordinates
    if sort_by_loc and batch_size < len(geometries):
        logger.error(f"Sorting the geometries with {len(geometries)} entries")
        geometries = sorted(geometries, key=lambda x: x.bounds[0])
        logger.error("Done sorting geometries")

    # TODO you could consider requesting a give number of points in the batch,
    # rather than number of objects.
    # TODO you could consider multiprocessing this since it's embarassingly parallel

    # Wrap the iteration in tqdm if requested, else just return it
    iteration_decorator = lambda x: (
        tqdm(x, desc=f"Computing batched unary union with batch size {batch_size}")
        if verbose
        else x
    )

    # Compute batched version
    batched_unions = []
    for i in iteration_decorator(
        range(0, len(geometries), batch_size),
    ):
        batch = geometries[i : i + batch_size]
        batched_unions.append(unary_union(batch, grid_size=grid_size))

    # Simplify the geometry to reduce the number of points
    # Don't do this if it would otherwise be returned as-is
    if simplify_tol > 0.0 and len(batched_unions) > 1:
        # Simplify then buffer, to make sure we don't have holes
        logger.info(
            f"Lengths before simplification {[len(bu.geoms) for bu in batched_unions]}"
        )
        batched_unions = [
            simplify(bu, simplify_tol).buffer(simplify_tol)
            for bu in tqdm(batched_unions, desc="simplifying polygons")
        ]
        logger.info(
            f"Lengths after simplification {[len(bu.geoms) for bu in batched_unions]}"
        )
    # Recurse this process until there's only one merged geometry
    # All future calls will use the subsequent_batch_size
    # TODO this batch size could be computed more inteligently, or sidestepped by requesting a number of points
    # Don't sort because this should already be sorted
    # Don't simplify because we don't want to repeatedly degrade the geometry
    return batched_unary_union(
        batched_unions,
        batch_size=subsequent_batch_size,
        grid_size=grid_size,
        subsequent_batch_size=subsequent_batch_size,
        sort_by_loc=False,
        simplify_tol=0.0,
    )


def get_scale_from_transform(transform: typing.Union[np.ndarray, None]):
    if transform is None:
        return 1

    if transform.shape != (4, 4):
        raise ValueError(f"Transform shape was {transform.shape}")

    transform_determinant = np.linalg.det(transform[:3, :3])
    scale_factor = np.cbrt(transform_determinant)
    return scale_factor


def unit_vector(vector):
    """Returns the unit vector of the vector."""
    return vector / np.linalg.norm(vector)


# https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
def angle_between(v1, v2):
    """Returns the angle in radians between vectors 'v1' and 'v2'::

    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966
    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0
    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def orthogonal_projection(v1, v2):
    # component of v2 along v1
    scalar = np.dot(v1, v2) / np.linalg.norm(v1) ** 2
    return scalar * v1


def projection_onto_plane(v1, e1, e2):
    # Compute the projection of vector v1 onto the plane defined by e1, e2
    plane_normal = np.cross(e1, e2)
    out_of_plane_projection = orthogonal_projection(plane_normal, v1)
    in_plane_projection = v1 - out_of_plane_projection
    return in_plane_projection



import argparse
from typing import Union

import numpy as np
import pyvista as pv
from scipy.sparse import load_npz, save_npz
from SetCoverPy.setcover import SetCover

from geograypher.cameras import MetashapeCameraSet, SegmentorPhotogrammetryCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshIndexPredictions
from geograypher.predictors.derived_segmentors import ImageIDSegmentor
from geograypher.utils.files import ensure_containing_folder


def determine_minimum_overlapping_images(
    mesh_file: PATH_TYPE,
    cameras_file: PATH_TYPE,
    image_folder: PATH_TYPE = "",
    compute_projection: bool = False,
    compute_minimal_set: bool = False,
    save_selected_images: bool = False,
    projections_filename: Union[PATH_TYPE, None] = None,
    selected_images_mask_filename: Union[PATH_TYPE, None] = None,
    selected_images_save_folder: Union[PATH_TYPE, None] = None,
    downsample_target: float = 1,
    min_observations_to_be_included: int = 1,
    vis: bool = False,
):
    """Determine a subset of images that together cover (or nearly cover) the entire scene.

    Args:
        mesh_file (PATH_TYPE):
            Path to the mesh file exported by Metashape
        cameras_file (PATH_TYPE):
            Path to the cameras file exported by Metashape
        image_folder (PATH_TYPE, optional):
            Path to the image folder used to generate the project. Only required if
            save_selected_images=True. Defaults to "".
        compute_projection (bool, optional):
            Compute which images project to which faces. Defaults to False.
        compute_minimal_set (bool, optional):
            Compute the minimal set of images required to cover the whole scene. Requires that the
            projections_filename points to a valid projections file. Defaults to False.
        save_selected_images (bool, optional):
            Save the selected images out to selected_images_save_folder. Requires that
            selected_images_mask_filename contain a mask inidicating which images should be
            included. Defaults to False.
        projections_filename (Union[PATH_TYPE, None], optional):
            Where to save or load from the projections to faces. Will have a .npz extension.
            Defaults to None.
        selected_images_mask_filename (Union[PATH_TYPE, None], optional):
            Where to save or load from the selected subset of images as a binary mask. Will have
            a .npy extension. Defaults to None.
        selected_images_save_folder (Union[PATH_TYPE, None], optional):
            Where to save the selected folder of images. Defaults to None.
        downsample_target (float, optional):
            Downsample the mesh to this fraction of the original faces. Defaults to 1.
        min_observations_to_be_included (int, optional):
            Ensure that a camera observes all faces that are observed by at least this many cameras.
            Setting to a higher value allows you to avoid including cameras only because they are
            the sole camera to observe a few faces. Defaults to 1.
        vis (bool, optional):
            Show intermediate results. Note that this will cause the process to hang until the
            visualization is closed. Defaults to False.
    """
    if compute_projection:
        # Load the mesh and optionally downsample it
        mesh = TexturedPhotogrammetryMeshIndexPredictions(
            mesh=mesh_file,
            downsample_target=downsample_target,
            transform_filename=cameras_file,
        )
        # Load the camera set
        camera_set = MetashapeCameraSet(
            camera_file=cameras_file, image_folder=image_folder
        )
        # Visualize the inputs if requested
        if vis:
            mesh.vis(camera_set=camera_set, frustum_scale=15)

        # Get all the image filenames
        image_filenames = camera_set.get_image_filename(index=None, absolute=True)
        # Create a segmentor that returns an image with all pixels set to the index of the image
        # within the set of images
        segmentor = ImageIDSegmentor(image_filenames=image_filenames)
        # Wrap the camera set in the segmentor
        segmentor_camera_set = SegmentorPhotogrammetryCameraSet(
            base_camera_set=camera_set, segmentor=segmentor
        )

        # Project the images to the mesh
        _, additional_info = mesh.aggregate_projected_images(
            cameras=segmentor_camera_set, n_classes=len(camera_set)
        )
        # Extract the summed (unnormalized projections)
        summed_projections = additional_info["summed_projections"]
        # Save the projections
        ensure_containing_folder(projections_filename)
        save_npz(projections_filename, summed_projections)

        if vis:
            # Compute the number of projections per face
            counts = np.squeeze(np.asarray(np.sum(summed_projections, axis=1)))
            # Compute the index of the camera that projects to a face. If multiple do, the one with
            # the lowest index will be reported
            projected_camera = np.asarray(np.argmax(summed_projections, axis=1)).astype(
                float
            )
            # Mask out faces with no projections
            projected_camera[counts == 0] = np.nan

            print("Showing the number of projections per face")
            mesh.vis(
                vis_scalars=counts,
                plotter_kwargs={"title": "Number of cameras projecting to each face"},
            )
            mesh.vis(
                vis_scalars=projected_camera,
                plotter_kwargs={
                    "title": "Index of projected camera, lowest in case of multiple"
                },
            )

    if compute_minimal_set:
        # Load the projections
        projection_matrix = load_npz(projections_filename).astype(bool)
        # Determine how many images project to each face
        projected_images_per_face = np.squeeze(
            np.asarray(np.sum(projection_matrix, axis=1))
        )
        # Filter out any faces that don't have at least the threshold observation. This stops us
        # from requiring an image just because it observes a few faces that no other image does
        valid_rows = np.squeeze(
            projected_images_per_face >= min_observations_to_be_included
        )
        projection_matrix = projection_matrix[valid_rows, :]
        # Convert to a dense matrix
        # TODO see if we can avoid this step since it takes a fair bit of time and memory
        projection_matrix = np.asarray(projection_matrix.todense())

        # Define the costs for including each image as unit
        set_costs = np.ones(projection_matrix.shape[1]).astype(int)

        # Set up the set cover problem
        problem = SetCover(projection_matrix, set_costs)
        # Solve the problem
        solution_cost, time_used = problem.SolveSCP()
        print(
            f"The solution cost is {solution_cost} and solving took {time_used} minutes"
        )

        # Save out the mask representing the selected images
        selected_images = problem.s
        ensure_containing_folder(selected_images_mask_filename)
        np.save(selected_images_mask_filename, selected_images)

    if save_selected_images:
        # Load the camera set
        camera_set = MetashapeCameraSet(
            camera_file=cameras_file, image_folder=image_folder
        )
        # Load the mask identifying the selected cameras
        cameras_mask = np.load(selected_images_mask_filename)
        # Convert it from a boolean mask to indices
        cameras_indices = np.where(cameras_mask)[0]
        # Take the corresponding set of cameras
        subset_cameras_set = camera_set.get_subset_cameras(cameras_indices)
        # Save out the images corresponding to the selected subset
        subset_cameras_set.save_images(output_folder=selected_images_save_folder)

        # Show the selected cameras bigger
        if vis:
            plotter = pv.Plotter()
            # TODO update these constants to be more robust
            # The issue is they are in the internal coordinate system of the camera set
            # and the scale changes between projects. Ideally, we'd want to specify it in meters.
            camera_set.vis(plotter=plotter, frustum_scale=0.4)
            subset_cameras_set.vis(plotter=plotter, frustum_scale=0.8)
            print("Showing the cameras, with selected ones larger.")
            plotter.show()


def parse_args():
    description = (
        "This script determines a minimum set of images that fully observe the mesh."
        + "The command line arguments are directly passed to"
        + "geograypher.entrypoints.workflow_functions.determine_minimum_overlapping_images "
        + "which has the following documentation:\n\n"
        + determine_minimum_overlapping_images.__doc__
    )
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description
    )
    parser.add_argument("--mesh-file")
    parser.add_argument("--cameras-file")
    parser.add_argument("--image-folder")
    parser.add_argument("--compute-projection", action="store_true")
    parser.add_argument("--compute-minimal-set", action="store_true")
    parser.add_argument("--save-selected-images", action="store_true")
    parser.add_argument("--projections-filename")
    parser.add_argument("--selected-images-save-folder")
    parser.add_argument("--downsample-target", default=1.0, type=float)
    parser.add_argument("--min-observations-to-be-included", default=1, type=float)
    parser.add_argument("--vis", action="store_true")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse command line args
    args = parse_args()
    # Pass command line args to determine_minimum_overlapping_images
    determine_minimum_overlapping_images(**args.__dict__)



import argparse
from pathlib import Path

from geograypher.utils.prediction_metrics import (
    compute_confusion_matrix_from_geospatial,
)


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--prediction-file",
        type=Path,
        required=True,
        help="Path to vector or raster prediction file",
    )
    parser.add_argument(
        "--groundtruth-file",
        type=Path,
        required=True,
        help="Path to vector or raster label file",
    )
    parser.add_argument("--class-names", nargs="+")
    parser.add_argument("--vis-savefile")
    parser.add_argument("--normalize", action="store_true")
    parser.add_argument("--column-name")
    parser.add_argument(
        "--metrics-output-file",
        type=Path,
        help="Write out metrics to this file in json format",
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()

    cf_matrix, classes, accuracy = compute_confusion_matrix_from_geospatial(
        prediction_file=args.prediction_file,
        groundtruth_file=args.groundtruth_file,
        class_names=args.class_names,
        vis_savefile=args.vis_savefile,
        normalize=args.normalize,
        column_name=args.column_name,
    )



import piexif
from PIL import Image


def get_GPS_exif(filename):
    im = Image.open(filename)
    exif_dict = piexif.load(im.info["exif"])
    lat = exif_dict["GPS"][piexif.GPSIFD.GPSLatitude]
    lon = exif_dict["GPS"][piexif.GPSIFD.GPSLongitude]
    lat = (
        lat[0][0] / lat[0][1]
        + lat[1][0] / (lat[1][1] * 60)
        + lat[2][0] / (lat[2][1] * 3600)
    )
    lon = (
        lon[0][0] / lon[0][1]
        + lon[1][0] / (lon[1][1] * 60)
        + lon[2][0] / (lon[2][1] * 3600)
    )
    # https://stackoverflow.com/questions/54196347/overwrite-gps-coordinates-in-image-exif-using-python-3-6
    return lon, lat



from pathlib import Path

from geograypher.constants import PATH_TYPE


def ensure_folder(folder: PATH_TYPE):
    """Ensure this folder, and parent folders, exist. Nothing happens if already present

    Args:
        folder (PATH_TYPE): Path to folder to ensure exists
    """
    folder = Path(folder)
    # Create this folder and all parent folders if needed. Nothing happens if it already exists
    folder.mkdir(parents=True, exist_ok=True)


def ensure_containing_folder(filename: PATH_TYPE):
    """Ensure the folder containing this file exists. Nothing happens if already present.

    Args:
        filename (PATH_TYPE): The path to the file for which the containing folder should be created
    """
    # Cast the file to a pathlib Path
    filename = Path(filename)
    # Get the folder above it
    containing_folder = filename.parent
    # Create this folder and all parent folders if needed. Nothing happens if it already exists
    ensure_folder(containing_folder)



import argparse

import numpy as np

from geograypher.cameras import MetashapeCameraSet
from geograypher.meshes import TexturedPhotogrammetryMesh


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--mesh-file")
    parser.add_argument("--camera-file")
    parser.add_argument("--texture")
    parser.add_argument("--downsample-target", type=float, default=1.0)
    parser.add_argument("--ROI")
    parser.add_argument("--ROI-buffer-meters", type=float, default=0.0)
    parser.add_argument("--convert-texture-to-max-class", action="store_true")

    args = parser.parse_args()
    return args


def visualize(
    mesh_file,
    camera_file,
    texture,
    downsample_target,
    ROI,
    ROI_buffer_meters,
    convert_texture_to_max_class,
):
    if convert_texture_to_max_class:
        texture = np.load(texture)
        max_class = np.argmax(texture, axis=1).astype(float)
        proj_per_face = np.sum(texture, axis=1)
        invalid_faces = np.logical_or(
            proj_per_face == 0, np.logical_not(np.isfinite(proj_per_face))
        )

        max_class[invalid_faces] = np.nan
        texture = max_class

    # Load the camera set if provided
    if camera_file is not None:
        camera_set = MetashapeCameraSet(camera_file=camera_file, image_folder="")
    else:
        camera_set = None

    mesh = TexturedPhotogrammetryMesh(
        mesh_file,
        ROI=ROI,
        ROI_buffer_meters=ROI_buffer_meters,
        downsample_target=downsample_target,
        texture=texture,
    )

    mesh.vis(camera_set=camera_set)


if __name__ == "__main__":
    args = parse_args()
    visualize(**args.__dict__)



import hashlib
import json
import logging
import os
import shutil
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Tuple, Union

import geopandas as gpd
import networkx
import numpy as np
import numpy.ma as ma
import pyvista as pv
from pyvista import demos
from scipy.spatial.distance import pdist
from shapely import MultiPolygon, Point, Polygon, unary_union
from skimage.io import imread
from skimage.transform import resize
from tqdm import tqdm

from geograypher.constants import (
    DEFAULT_FRUSTUM_SCALE,
    EARTH_CENTERED_EARTH_FIXED_CRS,
    EXAMPLE_INTRINSICS,
    LAT_LON_CRS,
    PATH_TYPE,
)
from geograypher.predictors.derived_segmentors import TabularRectangleSegmentor
from geograypher.utils.files import ensure_containing_folder
from geograypher.utils.geometric import (
    angle_between,
    get_scale_from_transform,
    orthogonal_projection,
    projection_onto_plane,
)
from geograypher.utils.geospatial import convert_CRS_3D_points, ensure_projected_CRS
from geograypher.utils.image import get_GPS_exif
from geograypher.utils.numeric import (
    compute_approximate_ray_intersection,
    triangulate_rays_lstsq,
)
from geograypher.utils.visualization import safe_start_xvfb


class PhotogrammetryCamera:
    def __init__(
        self,
        image_filename: PATH_TYPE,
        cam_to_world_transform: np.ndarray,
        f: float,
        cx: float,
        cy: float,
        image_width: int,
        image_height: int,
        distortion_params: Dict[str, float] = {},
        lon_lat: Union[None, Tuple[float, float]] = None,
        local_to_epsg_4978_transform: Union[np.array, None] = None,
    ):
        """Represents the information about one camera location/image as determined by photogrammetry

        Args:
            image_filename (PATH_TYPE): The image used for reconstruction
            transform (np.ndarray): A 4x4 transform representing the camera-to-world transform
            f (float): Focal length in pixels
            cx (float): Principle point x (pixels) from center
            cy (float): Principle point y (pixels) from center
            image_width (int): Input image width pixels
            image_height (int): Input image height pixels
            distortion_params (dict, optional): Distortion parameters, currently unused
            lon_lat (Union[None, Tuple[float, float]], optional): Location, defaults to None
        """
        self.image_filename = image_filename
        self.cam_to_world_transform = cam_to_world_transform
        self.world_to_cam_transform = np.linalg.inv(cam_to_world_transform)
        self.f = f
        self.cx = cx
        self.cy = cy
        self.image_width = image_width
        self.image_height = image_height
        self.distortion_params = distortion_params
        self.local_to_epsg_4978_transform = local_to_epsg_4978_transform

        if lon_lat is None:
            self.lon_lat = (None, None)
        else:
            self.lon_lat = lon_lat

        self.image_size = (image_height, image_width)
        self.image = None
        self.cache_image = (
            False  # Only set to true if you can hold all images in memory
        )

    def get_camera_hash(self, include_image_hash: bool = False):
        """Generates a hash value for the camera's geometry and optionally includes the image

        Args:
            include_image_hash (bool, optional): Whether to include the image filename in the hash computation. Defaults to false.

        Returns:
            int: A hash value representing the current state of the camera
        """
        # Geometric information of hash
        transform_hash = self.cam_to_world_transform.tolist()
        camera_settings = {
            "transform": transform_hash,
            "f": self.f,
            "cx": self.cx,
            "cy": self.cy,
            "image_width": self.image_width,
            "image_height": self.image_height,
            "distortion_params": self.distortion_params,
            "lon_lat": self.lon_lat,
        }

        # Include the image associated with the hash if specified
        if include_image_hash:
            camera_settings["image_filename"] = str(self.image_filename)

        camera_settings_data = json.dumps(camera_settings, sort_keys=True)
        hasher = hashlib.sha256()
        hasher.update(camera_settings_data.encode("utf-8"))

        return hasher.hexdigest()

    def get_camera_properties(self):
        """Returns the properties about a camera.

        Returns:
            dict: A dictionary containing the focal length, principal point coordinates,
                image height, image width, distortion parameters, and world_to_cam_transform.
        """
        camera_properties = {
            "focal_length": self.f,
            "principal_point_x": self.cx,
            "principal_point_y": self.cy,
            "image_height": self.image_height,
            "image_width": self.image_width,
            "distortion_params": self.distortion_params,
            "world_to_cam_transform": self.world_to_cam_transform,
        }
        return camera_properties

    def get_image(self, image_scale: float = 1.0) -> np.ndarray:
        # Check if the image is cached
        if self.image is None:
            image = imread(self.image_filename)
            if image.dtype == np.uint8:
                image = image / 255.0

            # Avoid unneccesary read if we have memory
            if self.cache_image:
                self.image = image
        else:
            image = self.image

        # Resizing is never cached, consider revisiting
        if image_scale != 1.0:
            image = resize(
                image,
                (int(image.shape[0] * image_scale), int(image.shape[1] * image_scale)),
            )

        return image

    def get_image_filename(self):
        return self.image_filename

    def get_image_size(self, image_scale=1.0):
        """Return image size, potentially scaled

        Args:
            image_scale (float, optional): How much to scale by. Defaults to 1.0.

        Returns:
            tuple[int]: (h, w) in pixels
        """
        # We should never have to deal with other cases if the reported size is accurate
        if self.image_size is not None:
            pass
        elif self.image is not None:
            self.image_size = self.image.shape[:2]
        else:
            image = self.get_image()
            self.image_size = image.shape[:2]

        return (
            int(self.image_size[0] * image_scale),
            int(self.image_size[1] * image_scale),
        )

    def get_lon_lat(self, negate_easting=True):
        """Return the lon, lat tuple, reading from exif metadata if neccessary"""
        if None in self.lon_lat:
            self.lon_lat = get_GPS_exif(self.image_filename)

            if negate_easting:
                self.lon_lat = (-self.lon_lat[0], self.lon_lat[1])

        return self.lon_lat

    def get_camera_location(self, get_z_coordinate: bool = False):
        """Returns a tuple of camera coordinates from the camera-to-world transfromation matrix.
        Args:
            get_z_coordinate (bool):
                Flag that user can set if they want z-coordinates. Defaults to False.
        Returns:
            Tuple[float, float (, float)]: tuple containing internal mesh coordinates of the camera
        """
        return (
            tuple(self.cam_to_world_transform[0:3, 3])
            if get_z_coordinate
            else tuple(self.cam_to_world_transform[0:2, 3])
        )

    def get_camera_view_angle(self, in_deg: bool = True) -> tuple:
        """Get the off-nadir pitch and yaw angles, computed geometrically from the photogrammtery result

        Args:
            in_deg (bool, optional): Return the angles in degrees rather than radians. Defaults to True.

        Returns:
            tuple: (pitch-from-nadir, yaw-from-nadir). Units are defined by in_deg parameter
        """
        # This is the origin, a point at one unit along the principal axis, a point one unit
        # up (-Y), and a point one unit right (+X)
        points_in_camera_frame = np.array(
            [[0, 0, 0, 1], [0, 0, 1, 1], [0, -1, 0, 1], [1, 0, 0, 1]]
        ).T

        # Transform the points first into the world frame and then into the earth-centered,
        # earth-fixed frame
        points_in_ECEF = (
            self.local_to_epsg_4978_transform
            @ self.cam_to_world_transform
            @ points_in_camera_frame
        )
        # Remove the homogenous coordinate and transpose
        points_in_ECEF = points_in_ECEF[:-1].T
        # Convert to shapely points
        points_in_ECEF = [Point(*point) for point in points_in_ECEF]
        # Convert to a dataframe
        points_in_ECEF = gpd.GeoDataFrame(
            geometry=points_in_ECEF, crs=EARTH_CENTERED_EARTH_FIXED_CRS
        )

        # Convert to lat lon
        points_in_lat_lon = points_in_ECEF.to_crs(LAT_LON_CRS)
        # Convert to a local projected CRS
        points_in_projected_CRS = ensure_projected_CRS(points_in_lat_lon)
        # Extract the geometry
        points_in_projected_CRS = np.array(
            [[p.x, p.y, p.z] for p in points_in_projected_CRS.geometry]
        )

        # Compute three vectors starting at the camera origin
        view_vector = points_in_projected_CRS[1] - points_in_projected_CRS[0]
        up_vector = points_in_projected_CRS[2] - points_in_projected_CRS[0]
        right_vector = points_in_projected_CRS[3] - points_in_projected_CRS[0]

        # The nadir vector points straight down
        NADIR_VEC = np.array([0, 0, -1])

        # For pitch, project the view vector onto the plane defined by the up vector and the nadir
        pitch_projection_view_vec = projection_onto_plane(
            view_vector, up_vector, NADIR_VEC
        )
        # For yaw, project the view vector onto the plane defined by the right vector and the nadir
        yaw_projection_view_vec = projection_onto_plane(
            view_vector, right_vector, NADIR_VEC
        )

        # Find the angle between these projected vectors and the nadir vector
        pitch_angle = angle_between(pitch_projection_view_vec, NADIR_VEC)
        yaw_angle = angle_between(yaw_projection_view_vec, NADIR_VEC)

        # Return in degrees if requested
        if in_deg:
            return (np.rad2deg(pitch_angle), np.rad2deg(yaw_angle))
        # Return in radians
        return (pitch_angle, yaw_angle)

    def check_projected_in_image(
        self, homogenous_image_coords: np.ndarray, image_size: Tuple[int, int]
    ):
        """Check if projected points are within the bound of the image and in front of camera

        Args:
            homogenous_image_coords (np.ndarray): The points after the application of K[R|t]. (3, n_points)
            image_size (Tuple[int, int]): The size of the image (width, height) in pixels

        Returns:
            np.ndarray: valid_points_bool, boolean array corresponding to which points were valid (n_points)
            np.ndarray: valid_image_space_points, float array of image-space coordinates for only valid points, (n_valid_points, 2)
        """
        img_width, image_height = image_size

        # Divide by the z coord to project onto the image plane
        image_space_points = homogenous_image_coords[:2] / homogenous_image_coords[2:3]
        # Transpose for convenience, (n_points, 3)
        image_space_points = image_space_points.T

        # We only want to consider points in front of the camera. Simple projection cannot tell
        # if a point is on the same ray behind the camera
        in_front_of_cam = homogenous_image_coords[2] > 0

        # Check that the point is projected within the image and is in front of the camera
        # Pytorch doesn't have a logical_and.reduce operator, so this is the equivilent using boolean multiplication
        valid_points_bool = (
            (image_space_points[:, 0] > 0)
            * (image_space_points[:, 1] > 0)
            * (image_space_points[:, 0] < img_width)
            * (image_space_points[:, 1] < image_height)
            * in_front_of_cam
        )

        # Extract the points that are valid
        valid_image_space_points = image_space_points[valid_points_bool, :].to(
            torch.int
        )
        # Return the boolean array
        valid_points_bool = valid_points_bool.cpu().numpy()
        valid_image_space_points = valid_image_space_points.cpu().numpy
        return valid_points_bool, valid_image_space_points

    def extract_colors(
        self, valid_bool: np.ndarray, valid_locs: np.ndarray, img: np.ndarray
    ):
        """_summary_

        Args:
            valid_bool (np.ndarray): (n_points,) boolean array cooresponding to valid points
            valid_locs (np.ndarray): (n_valid, 2) float array of image-space locations (x,y)
            img (np.ndarray): (h, w, n_channels) image to query from

        Returns:
            np.ma.array: (n_points, n_channels) One color per valid vertex. Points that were invalid are masked out
        """
        # Set up the data arrays
        colors_per_vertex = np.zeros((valid_bool.shape[0], img.shape[2]))
        mask = np.ones((valid_bool.shape[0], img.shape[2])).astype(bool)

        # Set the entries which are valid to false, meaning a valid entry in the masked array
        # TODO see if I can use valid_bool directly instead
        valid_inds = np.where(valid_bool)[0]
        mask[valid_inds, :] = False

        # Extract coordinates
        i_locs = valid_locs[:, 1]
        j_locs = valid_locs[:, 0]
        # Index based on the coordinates
        valid_color_samples = img[i_locs, j_locs, :]
        # Insert the valid samples into the array at the valid locations
        colors_per_vertex[valid_inds, :] = valid_color_samples
        # Convert to a masked array
        masked_color_per_vertex = ma.array(colors_per_vertex, mask=mask)
        return masked_color_per_vertex

    def project_mesh_verts(self, mesh_verts: np.ndarray, img: np.ndarray, device: str):
        """Get a color per vertex using only projective geometry, without considering occlusion or distortion

        Returns:
            np.ma.array: (n_points, n_channels) One color per valid vertex. Points that were invalid are masked out
        """
        # [R|t] matrix
        transform_3x4_world_to_cam = torch.Tensor(
            self.world_to_cam_transform[:3, :]
        ).to(device)
        K = torch.Tensor(
            [
                [self.f, 0, self.image_width / 2.0 + self.cx],
                [0, self.f, self.image_width + self.cy],
                [0, 0, 1],
            ],
            device=device,
        )
        # K[R|t], (3,4). Premultiplying these two matrices avoids doing two steps of projections with all points
        camera_matrix = K @ transform_3x4_world_to_cam

        # Add the extra dimension of ones for matrix multiplication
        homogenous_mesh_verts = torch.concatenate(
            (
                torch.Tensor(mesh_verts).to(device),
                torch.ones((mesh_verts.shape[0], 1)).to(device),
            ),
            axis=1,
        ).T

        # TODO review terminology
        homogenous_camera_points = camera_matrix @ homogenous_mesh_verts
        # Determine what points project onto the image and at what locations
        valid_bool, valid_locs = self.check_projected_in_image(
            projected_verts=homogenous_camera_points,
            image_size=(self.image_width, self.image_height),
        )
        # Extract corresponding colors from the image
        colors_per_vertex = self.extract_colors(valid_bool, valid_locs, img)

        return colors_per_vertex

    def get_pyvista_camera(self, focal_dist: float = 10) -> pv.Camera:
        """
        Get a pyvista camera at the location specified by photogrammetry.
        Note that there is no principle point and only the vertical field of view is set

        Args:
            focal_dist (float, optional): How far away from the camera the center point should be. Defaults to 10.

        Returns:
            pv.Camera: The pyvista camera from that viewpoint.
        """
        # Instantiate a new camera
        camera = pv.Camera()
        # Get the position as the translational part of the transform
        camera_position = self.cam_to_world_transform[:3, 3]
        # Get the look point by transforming a ray along the camera's Z axis into world
        # coordinates and then adding this to the location
        camera_look = camera_position + self.cam_to_world_transform[:3, :3] @ np.array(
            (0, 0, focal_dist)
        )
        # Get the up direction of the camera by finding which direction the -Y (image up) vector is transformed to
        camera_up = self.cam_to_world_transform[:3, :3] @ np.array((0, -1, 0))
        # Compute the vertical field of view
        vertical_FOV_angle = np.rad2deg(2 * np.arctan((self.image_height / 2) / self.f))

        # Set the values
        camera.focal_point = camera_look
        camera.position = camera_position
        camera.up = camera_up
        camera.view_angle = vertical_FOV_angle

        return camera

    def vis(self, plotter: pv.Plotter = None, frustum_scale: float = 0.1):
        """
        Visualize the camera as a frustum, at the appropriate translation and
        rotation and with the given focal length and aspect ratio.


        Args:
            plotter (pv.Plotter): The plotter to add the visualization to
            frustum_scale (float, optional): The length of the frustum in world units. Defaults to 0.5.
        """
        scaled_halfwidth = self.image_width / (self.f * 2)
        scaled_halfheight = self.image_height / (self.f * 2)

        scaled_cx = self.cx / self.f
        scaled_cy = self.cy / self.f

        right = scaled_cx + scaled_halfwidth
        left = scaled_cx - scaled_halfwidth
        top = scaled_cy + scaled_halfheight
        bottom = scaled_cy - scaled_halfheight

        vertices = (
            np.array(
                [
                    [0, 0, 0],
                    [
                        right,
                        top,
                        1,
                    ],
                    [
                        right,
                        bottom,
                        1,
                    ],
                    [
                        left,
                        bottom,
                        1,
                    ],
                    [
                        left,
                        top,
                        1,
                    ],
                ]
            ).T
            * frustum_scale
        )
        # Make the coordinates homogenous
        vertices = np.vstack((vertices, np.ones((1, 5))))

        # Project the vertices into the world cordinates
        projected_vertices = self.cam_to_world_transform @ vertices

        # Deal with the case where there is a scale transform
        if self.cam_to_world_transform[3, 3] != 1.0:
            projected_vertices /= self.cam_to_world_transform[3, 3]

        ## mesh faces
        faces = np.hstack(
            [
                [3, 0, 1, 2],  # side
                [3, 0, 2, 3],  # bottom
                [3, 0, 3, 4],  # side
                [3, 0, 4, 1],  # top
                [3, 1, 2, 3],  # endcap tiangle #1
                [3, 3, 4, 1],  # endcap tiangle #2
            ]
        )
        # All blue except the top (-Y) surface is red
        face_colors = np.array(
            [[0, 0, 1], [1, 0, 0], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1]]
        ).astype(float)

        # Create a mesh for the camera frustum
        frustum = pv.PolyData(projected_vertices[:3].T, faces)
        # Unsure exactly what's going on here, but it's required for it to be valid
        frustum.triangulate()
        # Show the mesh with the given face colors
        # TODO understand how this understands it's face vs. vertex colors? Simply by checking the number of values?
        plotter.add_mesh(frustum, scalars=face_colors, rgb=True)

    def cast_rays(self, pixel_coords_ij: np.ndarray, line_length: float = 10):
        """Compute rays eminating from the camera

        Args:
            image_coords (np.ndarray): (n,2) array of (i,j) pixel coordinates in the image
            line_length (float, optional): How long the lines are. Defaults to 10. #TODO allow an array of different values

        Returns:
            np.array: The projected vertices, TODO
        """
        # Transform from i, j to x, y
        pixel_coords_xy = np.flip(pixel_coords_ij, axis=1)

        # Cast a ray from the center of the mesh for vis
        principal_point = np.array(
            [[self.image_width / 2.0 + self.cx, self.image_height / 2.0 + self.cy]]
        )
        centered_pixel_coords = pixel_coords_xy - principal_point
        scaled_pixel_coords = centered_pixel_coords / self.f

        n_points = len(scaled_pixel_coords)

        if n_points == 0:
            return

        line_verts = [
            np.array(
                [
                    [0, 0, 0, 1],
                    [
                        point[0] * line_length,
                        point[1] * line_length,
                        line_length,
                        1,
                    ],
                ]
            )
            for point in scaled_pixel_coords
        ]
        line_verts = np.concatenate(line_verts, axis=0).T

        projected_vertices = self.cam_to_world_transform @ line_verts

        # Handle scale in transform
        if self.cam_to_world_transform[3, 3] != 1.0:
            projected_vertices /= self.cam_to_world_transform[3, 3]

        projected_vertices = projected_vertices[:3, :].T

        return projected_vertices

    def vis_rays(
        self, pixel_coords_ij: np.ndarray, plotter: pv.Plotter, line_length: float = 10
    ):
        """Show rays eminating from the camera

        Args:
            image_coords (np.ndarray): (n,2) array of (i,j) pixel coordinates in the image
            plotter (pv.Plotter): Plotter to use.
            line_length (float, optional): How long the lines are. Defaults to 10. #TODO allow an array of different values
        """
        # If there are no detections, just skip it
        if len(pixel_coords_ij) == 0:
            return

        projected_vertices = self.cast_rays(
            pixel_coords_ij=pixel_coords_ij, line_length=line_length
        )
        n_points = int(projected_vertices.shape[0] / 2)

        lines = np.vstack(
            (
                np.full(n_points, fill_value=2),
                np.arange(0, 2 * n_points, 2),
                np.arange(0, 2 * n_points, 2) + 1,
            )
        ).T

        mesh = pv.PolyData(projected_vertices.copy(), lines=lines.copy())
        plotter.add_mesh(mesh)


class PhotogrammetryCameraSet:
    def __init__(
        self,
        cameras: Union[None, PhotogrammetryCamera, List[PhotogrammetryCamera]] = None,
        cam_to_world_transforms: Union[None, List[np.ndarray]] = None,
        intrinsic_params_per_sensor_type: Dict[int, Dict[str, float]] = {
            0: EXAMPLE_INTRINSICS
        },
        image_filenames: Union[List[PATH_TYPE], None] = None,
        lon_lats: Union[None, List[Union[None, Tuple[float, float]]]] = None,
        image_folder: PATH_TYPE = None,
        sensor_IDs: List[int] = None,
        validate_images: bool = False,
        local_to_epsg_4978_transform: Union[np.array, None] = None,
    ):
        """_summary_

        Args:
            cam_to_world_transforms (List[np.ndarray]): The list of 4x4 camera to world transforms
            intrinsic_params_per_sensor (Dict[int, Dict]): A dictionary mapping from an int camera ID to the intrinsic parameters
            image_filenames (List[PATH_TYPE]): The list of image filenames, ideally absolute paths
            lon_lats (Union[None, List[Union[None, Tuple[float, float]]]]): A list of lon,lat tuples, or list of Nones, or None
            image_folder (PATH_TYPE): The top level folder of the images
            sensor_IDs (List[int]): The list of sensor IDs, that index into the sensors_params_dict
            validate_images (bool, optional): Should the existance of the images be checked. Defaults to False.

        Raises:
            ValueError: _description_
        """
        # Create an object using the supplied cameras
        if cameras is not None:
            if isinstance(cameras, PhotogrammetryCamera):
                cameras = [cameras]
            self.cameras = cameras
            return

        # Standardization
        n_transforms = len(cam_to_world_transforms)

        # Create list of Nones for image filenames if not set
        if image_filenames is None:
            image_filenames = [None] * n_transforms

        if sensor_IDs is None and len(intrinsic_params_per_sensor_type) == 1:
            # Create a list of the only index if not set
            sensor_IDs = [
                list(intrinsic_params_per_sensor_type.keys())[0]
            ] * n_transforms
        elif len(sensor_IDs) != n_transforms:
            raise ValueError(
                f"Number of sensor_IDs ({len(sensor_IDs)}) is different than the number of transforms ({n_transforms})"
            )

        # If lon lats is None, set it to a list of Nones per transform
        if lon_lats is None:
            lon_lats = [None] * n_transforms

        if image_folder is None:
            # TODO set it to the least common ancestor of all filenames
            pass

        # Record the values
        # TODO see if we ever use these
        self.cam_to_world_transforms = cam_to_world_transforms
        self.intrinsic_params_per_sensor_type = intrinsic_params_per_sensor_type
        self.image_filenames = image_filenames
        self.lon_lats = lon_lats
        self.sensor_IDs = sensor_IDs
        self.image_folder = image_folder

        if validate_images:
            missing_images, invalid_images = self.find_mising_images()
            if len(missing_images) > 0:
                print(f"Deleting {len(missing_images)} missing images")
                valid_images = np.where(np.logical_not(invalid_images))[0]
                self.image_filenames = np.array(self.image_filenames)[
                    valid_images
                ].tolist()
                # Avoid calling .tolist() because this will recursively set all elements to lists
                # when this should be a list of np.arrays
                self.cam_to_world_transforms = [
                    x for x in np.array(self.cam_to_world_transforms)[valid_images]
                ]
                self.sensor_IDs = np.array(self.sensor_IDs)[valid_images].tolist()
                self.lon_lats = np.array(self.lon_lats)[valid_images].tolist()

        self.cameras = []

        for image_filename, cam_to_world_transform, sensor_ID, lon_lat in zip(
            self.image_filenames,
            self.cam_to_world_transforms,
            self.sensor_IDs,
            self.lon_lats,
        ):
            sensor_params = self.intrinsic_params_per_sensor_type[sensor_ID]
            # This means the sensor did not have enough parameters to be valid
            if sensor_params is None:
                continue

            new_camera = PhotogrammetryCamera(
                image_filename,
                cam_to_world_transform,
                lon_lat=lon_lat,
                local_to_epsg_4978_transform=local_to_epsg_4978_transform,
                **sensor_params,
            )
            self.cameras.append(new_camera)

    def __len__(self):
        return self.n_cameras()

    def __getitem__(self, slice):
        subset_cameras = self.cameras[slice]
        if isinstance(subset_cameras, PhotogrammetryCamera):
            # this is just one item indexed
            return subset_cameras
        # else, wrap the list of cameras in a CameraSet
        return PhotogrammetryCameraSet(subset_cameras)

    def get_image_folder(self):
        return self.image_folder

    def find_mising_images(self):
        invalid_mask = []
        for image_file in self.image_filenames:
            if not image_file.is_file():
                invalid_mask.append(True)
            else:
                invalid_mask.append(False)
        invalid_images = np.array(self.image_filenames)[np.array(invalid_mask)].tolist()

        return invalid_images, invalid_mask

    def n_cameras(self) -> int:
        """Return the number of cameras"""
        return len(self.cameras)

    def n_image_channels(self) -> int:
        """Return the number of channels in the image"""
        return 3

    def get_cameras_in_folder(self, folder: PATH_TYPE):
        """Return the camera set with cameras corresponding to images in that folder

        Args:
            folder (PATH_TYPE): The folder location

        Returns:
            PhotogrammetryCameraSet: A copy of the camera set with only the cameras from that folder
        """
        # Get the inds where that camera is in the folder
        imgs_in_folder_inds = [
            i
            for i in range(len(self.cameras))
            if self.cameras[i].image_filename.is_relative_to(folder)
        ]
        # Return the PhotogrammetryCameraSet with those subset of cameras
        subset_cameras = self.get_subset_cameras(imgs_in_folder_inds)
        return subset_cameras

    def get_subset_cameras(self, inds: List[int]):
        subset_camera_set = deepcopy(self)
        subset_camera_set.cameras = [subset_camera_set[i] for i in inds]
        return subset_camera_set

    def get_image_by_index(self, index: int, image_scale: float = 1.0) -> np.ndarray:
        return self[index].get_image(image_scale=image_scale)

    def get_camera_view_angles(self, in_deg: bool = True) -> List[Tuple[float]]:
        """Compute the pitch and yaw off-nadir for all cameras in the set

        Args:
            in_deg (bool, optional): Return the angles in degrees rather than radians. Defaults to True.

        Returns:
            List[Tuple[float]]: A list of (pitch, yaw) tuples for each camera.
        """
        return [
            camera.get_camera_view_angle(in_deg=in_deg)
            for camera in tqdm(self.cameras, desc="Computing view angles")
        ]

    def get_image_filename(self, index: Union[int, None], absolute=True):
        """Get the image filename(s) based on the index

        Args:
            index (Union[int, None]):
                Return the filename of the camera at this index, or all filenames if None.
                #TODO update to support lists of integer indices as well
            absolute (bool, optional):
                Return the absolute filepath, as oposed to the path relative to the image folder.
                Defaults to True.

        Returns:
            typing.Union[PATH_TYPE, list[PATH_TYPE]]:
                If an integer index is provided, one path will be returned. If None, a list of paths
                will be returned.
        """
        if index is None:
            return [
                self.get_image_filename(i, absolute=absolute)
                for i in range(len(self.cameras))
            ]

        filename = self.cameras[index].get_image_filename()
        if absolute:
            return Path(filename)
        else:
            return Path(filename).relative_to(self.get_image_folder())

    def save_images(self, output_folder, copy=False, remove_folder: bool = True):
        if remove_folder:
            if os.path.isdir(output_folder):
                print(f"about to remove {output_folder}")
                shutil.rmtree(output_folder)

        for i in tqdm(
            range(len(self.cameras)),
            f"{'copying' if copy else 'linking'} images to {output_folder}",
        ):
            output_file = Path(
                output_folder, self.get_image_filename(i, absolute=False)
            )
            ensure_containing_folder(output_file)
            src_file = self.get_image_filename(i, absolute=True)
            if copy:
                try:
                    shutil.copy(src_file, output_file)
                except FileNotFoundError:
                    logging.warning(f"Could not find {src_file}")
            else:
                os.symlink(src_file, output_file)

    def get_lon_lat_coords(self):
        """Returns a list of GPS coords for each camera"""
        return [x.get_lon_lat() for x in self.cameras]

    def get_camera_locations(self, **kwargs):
        """
        Returns a list of camera locations for each camera.

        Args:
            **kwargs: Keyword arguments to be passed to the PhotogrammetryCamera.get_camera_location method.

        Returns:
            List[Tuple[float, float] or Tuple[float, float, float]]:
                List of tuples containing the camera locations.
        """
        return [x.get_camera_location(**kwargs) for x in self.cameras]

    def get_subset_ROI(
        self,
        ROI: Union[PATH_TYPE, gpd.GeoDataFrame, Polygon, MultiPolygon],
        buffer_radius: float = 0,
        is_geospatial: bool = None,
    ):
        """Return cameras that are within a radius of the provided geometry

        Args:
            geodata (Union[PATH_TYPE, gpd.GeoDataFrame, Polygon, MultiPolygon]):
                This can be a Geopandas dataframe, path to a geofile readable by geopandas, or
                Shapely Polygon/MultiPolygon information that can be loaded into a geodataframe
            buffer_radius (float, optional):
                Return points within this buffer of the geometry. Defaults to 0. Represents
                meters if ROI is geospatial.
            is_geospatial (bool, optional):
                A flag for user to indicate if ROI is geospatial or not; if no flag is provided,
                the flag is set if the provided geodata has a CRS.
        Returns:
            subset_camera_set (List[PhotogrammetryCamera]):
                List of cameras that fall within the provided ROI
        """
        # construct GeoDataFrame if not provided
        if isinstance(ROI, (Polygon, MultiPolygon)):
            # assume geodata is lat/lon if is_geospatial is True
            if is_geospatial:
                ROI = gpd.GeoDataFrame(crs=LAT_LON_CRS, geometry=[ROI])
            else:
                ROI = gpd.GeoDataFrame(geometry=[ROI])
        elif not isinstance(ROI, gpd.GeoDataFrame):
            # Read in the geofile
            ROI = gpd.read_file(ROI)

        if is_geospatial is None:
            is_geospatial = ROI.crs is not None

        if not is_geospatial:
            # get internal coordinate system camera locations
            image_locations = [Point(*x) for x in self.get_camera_locations()]
            image_locations_df = gpd.GeoDataFrame(geometry=image_locations)
        else:
            # Make sure it's a geometric (meters-based) CRS
            ROI = ensure_projected_CRS(ROI)
            # Read the locations of all the points
            image_locations = [Point(*x) for x in self.get_lon_lat_coords()]
            # Create a dataframe, assuming inputs are lat lon
            image_locations_df = gpd.GeoDataFrame(
                geometry=image_locations, crs=LAT_LON_CRS
            )
            image_locations_df.to_crs(ROI.crs, inplace=True)

        # Drop all the fields except the geometry for computational reasons
        ROI = ROI["geometry"]
        # Buffer out by the requested distance
        ROI = ROI.buffer(buffer_radius)
        # Merge the potentially-individual polygons to one
        # TODO Do experiments to see if this step should be before or after the buffer.
        # Counterintuitively, it seems that after is faster
        ROI = unary_union(ROI.tolist())

        # Determine the binary mask for which cameras are within the ROI
        cameras_in_ROI = image_locations_df.within(ROI).to_numpy()
        # Convert to the integer indices
        cameras_in_ROI = np.where(cameras_in_ROI)[0]
        # Get the corresponding subset of cameras
        subset_camera_set = self.get_subset_cameras(cameras_in_ROI)
        return subset_camera_set

    def triangulate_detections(
        self,
        detector: TabularRectangleSegmentor,
        transform_to_epsg_4978=None,
        similarity_threshold_meters: float = 0.1,
        louvain_resolution: float = 2,
        vis: bool = True,
        plotter: pv.Plotter = pv.Plotter(),
        vis_ray_length_meters: float = 200,
    ) -> np.ndarray:
        """Take per-image detections and triangulate them to 3D locations

        Args:
            detector (TabularRectangleSegmentor):
                Produces detections per image using the get_detection_centers method
            transform_to_epsg_4978 (typing.Union[np.ndarray, None], optional):
                The 4x4 transform to earth centered earth fixed coordinates. Defaults to None.
            similarity_threshold_meters (float, optional):
                Consider rays a potential match if the distance between them is less than this
                value. Defaults to 0.1.
            louvain_resolution (float, optional):
                The resolution parameter of the networkx.louvain_communities function. Defaults to
                2.0.
            vis (bool, optional):
                Whether to show the detection projections and intersecting points. Defaults to True.
            plotter (pv.Plotter, optional):
                The plotter to add the visualizations to is vis=True. If not set, a new one will be
                created. Defaults to pv.Plotter().
            vis_ray_length_meters (float, optional):
                The length of the visualized rays in meters. Defaults to 200.

        Returns:
            np.ndarray:
                (n unique objects, 3), the 3D locations of the identified objects.
                If transform_to_epsg_4978 is set, then this is in (lat, lon, alt), if not, it's in the
                local coordinate system of the mesh
        """
        # Determine scale factor relating meters to internal coordinates
        meters_to_local_scale = 1 / get_scale_from_transform(transform_to_epsg_4978)
        similarity_threshold_local = similarity_threshold_meters * meters_to_local_scale
        vis_ray_length_local = vis_ray_length_meters * meters_to_local_scale

        # Record the lines corresponding to each detection and the associated image ID
        all_line_segments = []
        all_image_IDs = []

        # Iterate over the cameras
        for camera_ind in range(len(self)):
            # Get the image filename
            image_filename = str(self.get_image_filename(camera_ind, absolute=False))
            # Get the centers of associated detection from the detector
            # TODO, this only works with "detectors" that can look up the detections based on the
            # filename alone. In the future we might want to support real detectors that actually
            # use the image.
            detection_centers_pixels = detector.get_detection_centers(image_filename)
            # Get the individual camera
            camera = self[camera_ind]
            # Project rays given the locations of the detections in pixel coordinates
            line_segments = camera.cast_rays(
                pixel_coords_ij=detection_centers_pixels,
                line_length=vis_ray_length_local,
            )
            # If there are no detections, this will be None
            if line_segments is not None:
                # Record the line segments, which will be ordered as alternating (start, end) rows
                all_line_segments.append(line_segments)
                # Record which image ID generated each line
                all_image_IDs.append(
                    np.full(int(line_segments.shape[0] / 2), fill_value=camera_ind)
                )

        # Concatenate the lists of arrays into a single array
        all_line_segments = np.concatenate(all_line_segments, axis=0)
        all_image_IDs = np.concatenate(all_image_IDs, axis=0)

        # Get the starts and ends, which are alternating rows
        ray_starts = all_line_segments[0::2]
        segment_ends = all_line_segments[1::2]
        # Determine the direction
        ray_directions = segment_ends - ray_starts
        # Make the ray directions unit length
        ray_directions = ray_directions / np.linalg.norm(
            ray_directions, axis=1, keepdims=True
        )

        # Compute the distance matrix of ray-ray intersections
        num_dets = ray_starts.shape[0]
        interesection_dists = np.full((num_dets, num_dets), fill_value=np.nan)

        # Calculate the upper triangular matrix of ray-ray interesections
        for i in tqdm(range(num_dets), desc="Calculating quality of ray intersections"):
            for j in range(i, num_dets):
                # Extract starts and directions
                A = ray_starts[i]
                B = ray_starts[j]
                a = ray_directions[i]
                b = ray_directions[j]
                # TODO explore whether this could be vectorized
                dist, valid = compute_approximate_ray_intersection(A, a, B, b)

                interesection_dists[i, j] = dist if valid else np.nan

        # Filter out intersections that are above the threshold distance
        interesection_dists[interesection_dists > similarity_threshold_local] = np.nan

        # Determine which intersections are valid, represented by finite values
        i_inds, j_inds = np.where(np.isfinite(interesection_dists))

        # Build a list of (i, j, info_dict) tuples encoding the valid edges and their intersection
        # distance
        positive_edges = [
            (i, j, {"weight": 1 / interesection_dists[i, j]})
            for i, j in zip(i_inds, j_inds)
        ]

        # Build a networkx graph. The nodes represent an individual detection while the edges
        # represent the quality of the matches between detections.
        graph = networkx.Graph(positive_edges)
        # Determine Louvain communities which are sets of nodes. Ideally, this represents a set of
        # detections that all coorespond to one 3D object
        communities = networkx.community.louvain_communities(
            graph, weight="weight", resolution=louvain_resolution
        )
        # Sort the communities by size
        communities = sorted(communities, key=len, reverse=True)

        ## Triangulate the rays for each community to identify the 3D location
        community_points = []
        # Record the community IDs per detection
        community_IDs = np.full(num_dets, fill_value=np.nan)
        # Iterate over communities
        for community_ID, community in enumerate(communities):
            # Get the indices of the detections for that community
            community_detection_inds = np.array(list(community))
            # Record the community ID for the corresponding detection IDs
            community_IDs[community_detection_inds] = community_ID

            # Get the set of starts and directions for that community
            community_starts = ray_starts[community_detection_inds]
            community_directions = ray_directions[community_detection_inds]

            # Determine the least squares triangulation of the rays
            community_3D_point = triangulate_rays_lstsq(
                community_starts, community_directions
            )
            community_points.append(community_3D_point)

        # Stack all of the points into one vector
        community_points = np.vstack(community_points)

        # Show the rays and detections
        if vis:
            # Show the line segements
            # TODO: consider coloring these lines by community
            lines_mesh = pv.line_segments_from_points(all_line_segments)
            plotter.add_mesh(
                lines_mesh,
                scalars=community_IDs,
                label="Rays, colored by community ID",
            )
            # Show the triangulated communtities as red spheres
            detected_points = pv.PolyData(community_points)
            plotter.add_points(
                detected_points,
                color="r",
                render_points_as_spheres=True,
                point_size=10,
                label="Triangulated locations",
            )
            plotter.add_legend()

        # Convert the intersection points from the local mesh coordinate system to lat lon
        if transform_to_epsg_4978 is not None:
            # Append a column of all ones to make the homogenous coordinates
            community_points_homogenous = np.concatenate(
                [community_points, np.ones_like(community_points[:, 0:1])], axis=1
            )
            # Use the transform matrix to transform the points into the earth centered, earth fixed
            # frame, EPSG:4978
            community_points_epsg_4978 = (
                transform_to_epsg_4978 @ community_points_homogenous.T
            ).T
            # Convert the points from earth centered, earth fixed frame to lat lon
            community_points_lat_lon = convert_CRS_3D_points(
                community_points_epsg_4978,
                input_CRS=EARTH_CENTERED_EARTH_FIXED_CRS,
                output_CRS=LAT_LON_CRS,
            )
            # Set the community points to lat lon
            community_points = community_points_lat_lon

        # Return the 3D locations of the community points
        return community_points

    def vis(
        self,
        plotter: pv.Plotter = None,
        add_orientation_cube: bool = False,
        show: bool = False,
        frustum_scale: float = None,
        force_xvfb: bool = False,
        interactive_jupyter: bool = False,
    ):
        """Visualize all the cameras

        Args:
            plotter (pv.Plotter):
                Plotter to add the cameras to. If None, will be created and then plotted
            add_orientation_cube (bool, optional):
                Add a cube to visualize the coordinate system. Defaults to False.
            show (bool, optional):
                Show the results instead of waiting for other content to be added
            frustum_scale (float, optional):
                Size of cameras in world units. If None, will set to 1/120th of the maximum distance
                between two cameras.
            force_xvfb (bool, optional):
                Force a headless rendering backend
            interactive_jupyter (bool, optional):
                Will allow you to interact with the visualization in your notebook if supported by
                the notebook server. Otherwise will fail. Only applicable if `show=True`. Defaults
                to False.

        """

        if plotter is None:
            plotter = pv.Plotter()
            show = True

        # Determine pairwise distance between each camera and set frustum_scale to 1/120th of the maximum distance found
        if frustum_scale is None:
            if self.n_cameras() >= 2:
                camera_translation_matrices = np.array(
                    [transform[:3, 3] for transform in self.cam_to_world_transforms]
                )
                distances = pdist(camera_translation_matrices, metric="euclidean")
                max_distance = np.max(distances)
                frustum_scale = (
                    (max_distance / 120) if max_distance > 0 else DEFAULT_FRUSTUM_SCALE
                )
            # else, set it to a default
            else:
                frustum_scale = DEFAULT_FRUSTUM_SCALE

        for camera in self.cameras:
            camera.vis(plotter, frustum_scale=frustum_scale)
        if add_orientation_cube:
            # TODO Consider adding to a freestanding vis module
            ocube = demos.orientation_cube()
            plotter.add_mesh(ocube["cube"], show_edges=True)
            plotter.add_mesh(ocube["x_p"], color="blue")
            plotter.add_mesh(ocube["x_n"], color="blue")
            plotter.add_mesh(ocube["y_p"], color="green")
            plotter.add_mesh(ocube["y_n"], color="green")
            plotter.add_mesh(ocube["z_p"], color="red")
            plotter.add_mesh(ocube["z_n"], color="red")
            plotter.show_axes()

        if show:
            if force_xvfb:
                safe_start_xvfb()
            plotter.show(jupyter_backend="trame" if interactive_jupyter else "static")
